import json
import sys
import unittest
from datetime import datetime
from unittest.mock import patch, MagicMock

from keepercommander.error import CommandError
import keepercommander.vault as vault

def create_mock_params_and_record(record_type='pamUser'):
    mock_params = MagicMock()
    mock_params.rest_context.server_key_id = 8
    mock_params.session_token = 'base64_encoded_session_token'  # Mock a base64 encoded session token
    mock_params.record_cache = {'record_uid': MagicMock(record_type='pamUser')}
    mock_params.subfolder_record_cache = {'folder_uid': ['record_uid']}
    mock_params.folder_cache = {'folder_uid': MagicMock()}
    mock_params.record_rotation_cache = {
        'record_uid': {
            'pwd_complexity': 'eyJ0eXBlIjogInBhc3N3b3JkX2NvbXBsZXhpdHkiLCAidmFsdWUiOiAiY29tcGxleGl0eV92YWx1ZSJ9',
            'configuration_uid': 'config_uid',
            'schedule': '[]',
            'resourceUid': 'resource_uid',
            'revision': 1  # Ensure revision is set
        }
    }
    mock_params.rest_context.server_base = 'https://fake.keepersecurity.com'  # Mock URL as string

    mock_typed_record = MagicMock(spec=vault.TypedRecord)
    mock_typed_record.record_type = record_type
    mock_typed_record.record_uid = 'record_uid'
    mock_typed_record.title = 'Mock Title'  # Add the title attribute
    mock_typed_record.record_key = b'\x00' * 16  # Add the record_key attribute

    return mock_params, mock_typed_record


def create_mock_params():
    mock_params = MagicMock()
    mock_params.rest_context.server_key_id = 8
    mock_params.session_token = 'base64_encoded_session_token'
    mock_params.record_cache = {
        'record_uid': {
            'data_unencrypted': json.dumps({'title': 'Mock Title', 'type': 'pamMachine'})
        }
    }
    mock_params.subfolder_record_cache = {'folder_uid': ['record_uid']}
    mock_params.folder_cache = {'folder_uid': MagicMock()}
    mock_params.rest_context.server_base = 'https://fake.keepersecurity.com'

    return mock_params


if sys.version_info >= (3, 8):

    import requests
    from cryptography.hazmat.primitives.asymmetric import ec
    from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

    from keepercommander import crypto, utils
    from keepercommander.commands.discoveryrotation import (PAMCreateRecordRotationCommand, PAMListRecordRotationCommand,
                                                            PAMGatewayListCommand)

    class TestPAMCreateRecordRotationCommand(unittest.TestCase):

        def setUp(self):
            self.command = PAMCreateRecordRotationCommand()
            self.parser = self.command.get_parser()
            self.transmission_key = b'transmission_key'
            self.session_token = b'encrypted_session_token'
            self.private_key = ec.generate_private_key(ec.SECP256R1())
            self.public_key = self.private_key.public_key()

            # Serialize and deserialize the public key to ensure compatibility
            public_key_bytes = self.public_key.public_bytes(Encoding.X962, PublicFormat.UncompressedPoint)
            loaded_public_key = ec.EllipticCurvePublicKey.from_encoded_point(ec.SECP256R1(), public_key_bytes)

            self.encrypted_transmission_key = crypto.encrypt_ec(self.transmission_key, loaded_public_key)
            self.encrypted_session_token = crypto.encrypt_aes_v2(self.session_token, self.transmission_key)

        def test_parser(self):
            args = self.parser.parse_args(['--record', 'record_uid', '--force'])
            self.assertEqual(args.record_name, 'record_uid')
            self.assertTrue(args.force)

        @patch('keepercommander.vault.KeeperRecord.load')
        @patch('keepercommander.commands.discoveryrotation.TunnelDAG')
        @patch('keepercommander.rest_api.SERVER_PUBLIC_KEYS', {8: ec.generate_private_key(ec.SECP256R1()).public_key()})
        def test_execute_with_folder(self, mock_TunnelDAG, mock_load):
            mock_params, mock_typed_record = create_mock_params_and_record()

            mock_load.return_value = mock_typed_record

            mock_dag_instance = mock_TunnelDAG.return_value
            mock_dag_instance.linking_dag.has_graph = True
            mock_dag_instance.check_if_resource_has_admin.return_value = True
            mock_dag_instance.get_all_owners.return_value = ['resource_uid']
            mock_dag_instance.resource_belongs_to_config.return_value = True
            mock_dag_instance.user_belongs_to_resource.return_value = True
            mock_dag_instance.record.record_uid = 'config_uid'  # Ensure it returns a string

            kwargs = {
                'folder_name': 'folder_uid',
                'force': True  # Add force to the kwargs
            }

            self.command.execute(mock_params, **kwargs)
            self.assertTrue(mock_load.called)
            self.assertTrue(mock_TunnelDAG.called)

        @patch('keepercommander.vault.KeeperRecord.load', return_value=None)
        @patch('keepercommander.commands.discoveryrotation.TunnelDAG')
        @patch('keepercommander.rest_api.SERVER_PUBLIC_KEYS', {8: ec.generate_private_key(ec.SECP256R1()).public_key()})
        def test_execute_with_no_record(self, mock_TunnelDAG, mock_load):
            mock_params, _ = create_mock_params_and_record()
            mock_params.record_cache = {}

            kwargs = {
                'record_name': 'non_existent_record',
                'force': True  # Add force to the kwargs
            }

            with self.assertRaises(CommandError):
                self.command.execute(mock_params, **kwargs)

        @patch('keepercommander.vault.KeeperRecord.load', return_value=None)
        @patch('keepercommander.commands.discoveryrotation.TunnelDAG')
        @patch('keepercommander.rest_api.SERVER_PUBLIC_KEYS', {8: ec.generate_private_key(ec.SECP256R1()).public_key()})
        def test_execute_with_invalid_password_complexity(self, mock_TunnelDAG, mock_load):
            mock_params, _ = create_mock_params_and_record()

            kwargs = {
                'record_name': 'record_uid',
                'pwd_complexity': 'invalid_complexity',
                'force': True  # Add force to the kwargs
            }

            with self.assertRaises(CommandError):
                self.command.execute(mock_params, **kwargs)

        @patch('keepercommander.vault.KeeperRecord.load')
        @patch('keepercommander.commands.discoveryrotation.TunnelDAG')
        @patch('keepercommander.rest_api.SERVER_PUBLIC_KEYS', {8: ec.generate_private_key(ec.SECP256R1()).public_key()})
        def test_execute_with_valid_password_complexity(self, mock_TunnelDAG, mock_load):
            mock_params, mock_typed_record = create_mock_params_and_record()

            mock_load.return_value = mock_typed_record

            mock_dag_instance = mock_TunnelDAG.return_value
            mock_dag_instance.linking_dag.has_graph = True
            mock_dag_instance.check_if_resource_has_admin.return_value = True
            mock_dag_instance.get_all_owners.return_value = ['resource_uid']
            mock_dag_instance.resource_belongs_to_config.return_value = True
            mock_dag_instance.user_belongs_to_resource.return_value = True
            mock_dag_instance.record.record_uid = 'config_uid'  # Ensure it returns a string

            kwargs = {
                'record_name': 'record_uid',
                'pwd_complexity': '32,5,5,5,5',
                'force': True  # Add force to the kwargs
            }

            self.command.execute(mock_params, **kwargs)
            self.assertTrue(mock_load.called)
            self.assertEqual(mock_typed_record.record_key, b'\x00' * 16)

        @patch('keepercommander.vault.KeeperRecord.load')
        @patch('keepercommander.commands.discoveryrotation.TunnelDAG')
        @patch('keepercommander.rest_api.SERVER_PUBLIC_KEYS', {8: ec.generate_private_key(ec.SECP256R1()).public_key()})
        def test_execute_with_valid_record(self, mock_TunnelDAG, mock_load):
            mock_params, mock_typed_record = create_mock_params_and_record()

            mock_load.return_value = mock_typed_record

            mock_dag_instance = mock_TunnelDAG.return_value
            mock_dag_instance.linking_dag.has_graph = True
            mock_dag_instance.check_if_resource_has_admin.return_value = True
            mock_dag_instance.get_all_owners.return_value = ['resource_uid']
            mock_dag_instance.resource_belongs_to_config.return_value = True
            mock_dag_instance.user_belongs_to_resource.return_value = True
            mock_dag_instance.record.record_uid = 'config_uid'  # Ensure it returns a string

            kwargs = {
                'record_name': 'record_uid',
                'force': True  # Add force to the kwargs
            }

            self.command.execute(mock_params, **kwargs)
            self.assertTrue(mock_load.called)
            self.assertTrue(mock_TunnelDAG.called)
            self.assertEqual(mock_typed_record.record_key, b'\x00' * 16)


    class TestPAMResourceRotateCommand(unittest.TestCase):

        def setUp(self):
            self.command = PAMCreateRecordRotationCommand()
            self.parser = self.command.get_parser()

        def test_parser(self):
            args = self.parser.parse_args(['--record', "abcdefg", '--enable'])
            self.assertEqual(args.record_name, 'abcdefg')
            self.assertTrue(args.enable)

        @patch('keepercommander.vault_extensions.find_records')
        @patch('keepercommander.vault.KeeperRecord.load')
        @patch('keepercommander.commands.discoveryrotation.get_keeper_tokens')
        @patch('keepercommander.commands.discoveryrotation.TunnelDAG')
        def test_execute_with_enable(self, mock_tunneldag, mock_get_keeper_tokens, mock_load, mock_find_records):
            mock_dag_instance = mock_tunneldag.return_value
            mock_dag_instance.linking_dag.has_graph = True
            mock_dag_instance.resource_belongs_to_config.return_value = True

            mock_get_keeper_tokens.return_value = (b'token', b'encrypted_key', b'transmission_key')

            mock_params, mock_typed_record = create_mock_params_and_record('pamMachine')
            mock_load.return_value = mock_typed_record

            mock_pam_config_record = MagicMock(spec=vault.TypedRecord)
            mock_pam_config_record.record_uid = 'config_uid'
            mock_pam_config_record.record_type = 'pamConfiguration'  # Use a valid PAM configuration record type
            mock_find_records.return_value = [mock_pam_config_record]

            kwargs = {
                'record_name': 'record_uid',
                'enable': True,
                'config_uid': 'config_uid'
            }

            self.command.execute(mock_params, **kwargs)
            self.assertTrue(mock_load.called)
            self.assertTrue(mock_tunneldag.called)
            self.assertTrue(mock_get_keeper_tokens.called)

        @patch('keepercommander.vault.KeeperRecord.load', return_value=None)
        def test_execute_with_invalid_uid(self, mock_load):
            mock_params, _ = create_mock_params_and_record('pamMachine')

            kwargs = {
                'record_name': 'invalid_uid',
                'enable': True
            }

            with self.assertRaises(CommandError):
                self.command.execute(mock_params, **kwargs)

        @patch('keepercommander.vault.KeeperRecord.load')
        def test_execute_with_invalid_record_type(self, mock_load):
            mock_params, mock_typed_record = create_mock_params_and_record(record_type='invalid_type')
            mock_load.return_value = mock_typed_record

            kwargs = {
                'record_name': 'record_uid',
                'enable': True
            }

            with self.assertRaises(CommandError):
                self.command.execute(mock_params, **kwargs)

        @patch('keepercommander.vault_extensions.find_records')
        @patch('keepercommander.vault.KeeperRecord.load')
        @patch('keepercommander.commands.discoveryrotation.get_keeper_tokens')
        @patch('keepercommander.commands.discoveryrotation.TunnelDAG')
        def test_execute_with_disable(self, mock_tunneldag, mock_get_keeper_tokens, mock_load, mock_find_records):
            mock_dag_instance = mock_tunneldag.return_value
            mock_dag_instance.linking_dag.has_graph = True
            mock_dag_instance.resource_belongs_to_config.return_value = True

            mock_get_keeper_tokens.return_value = (b'token', b'encrypted_key', b'transmission_key')

            mock_params, mock_typed_record = create_mock_params_and_record('pamMachine')
            mock_load.return_value = mock_typed_record

            mock_pam_config_record = MagicMock(spec=vault.TypedRecord)
            mock_pam_config_record.record_uid = 'config_uid'
            mock_pam_config_record.record_type = 'pamConfiguration'  # Use a valid PAM configuration record type
            mock_find_records.return_value = [mock_pam_config_record]

            kwargs = {
                    'record_name': 'record_uid',
                    'disable': True,
                    'config_uid': 'config_uid'
                }

            self.command.execute(mock_params, **kwargs)
            self.assertTrue(mock_load.called)
            self.assertTrue(mock_tunneldag.called)
            self.assertTrue(mock_get_keeper_tokens.called)

        @patch('keepercommander.vault_extensions.find_records')
        @patch('keepercommander.vault.KeeperRecord.load')
        @patch('keepercommander.commands.discoveryrotation.get_keeper_tokens')
        @patch('keepercommander.commands.discoveryrotation.TunnelDAG')
        def test_execute_with_enable_and_admin(self, mock_tunneldag, mock_get_keeper_tokens, mock_load, mock_find_records):
            mock_dag_instance = mock_tunneldag.return_value
            mock_dag_instance.linking_dag.has_graph = True
            mock_dag_instance.resource_belongs_to_config.return_value = True

            mock_get_keeper_tokens.return_value = (b'token', b'encrypted_key', b'transmission_key')

            mock_params, mock_typed_record = create_mock_params_and_record('pamMachine')
            mock_load.return_value = mock_typed_record

            mock_pam_config_record = MagicMock(spec=vault.TypedRecord)
            mock_pam_config_record.record_uid = 'config_uid'
            mock_pam_config_record.record_type = 'pamConfiguration'  # Use a valid PAM configuration record type
            mock_find_records.return_value = [mock_pam_config_record]

            kwargs = {
                'record_name': 'record_uid',
                'enable': True,
                'config_uid': 'config_uid',
                'admin': 'admin_uid'
            }

            self.command.execute(mock_params, **kwargs)
            self.assertTrue(mock_load.called)
            self.assertTrue(mock_tunneldag.called)
            self.assertTrue(mock_get_keeper_tokens.called)
            mock_dag_instance.link_user_to_resource.assert_called_with('admin_uid', 'record_uid', is_admin=True)


    class TestPAMListRecordRotationCommand(unittest.TestCase):

        def setUp(self):
            self.command = PAMListRecordRotationCommand()
            self.parser = self.command.get_parser()

        def test_parser(self):
            args = self.parser.parse_args(['--verbose'])
            self.assertTrue(args.is_verbose)

        @patch('keepercommander.commands.discoveryrotation.router_get_rotation_schedules')
        @patch('keepercommander.commands.discoveryrotation.router_get_connected_gateways')
        @patch('keepercommander.commands.discoveryrotation.pam_configurations_get_all')
        @patch('keepercommander.commands.discoveryrotation.gateway_helper.get_all_gateways')
        @patch('keepercommander.commands.discoveryrotation.pam_decrypt_configuration_data')
        @patch('keepercommander.commands.discoveryrotation.dump_report_data')
        def test_execute(self, mock_dump_report_data, mock_pam_decrypt_configuration_data, mock_get_all_gateways,
                         mock_pam_configurations_get_all, mock_router_get_connected_gateways, mock_router_get_rotation_schedules):
            mock_params = create_mock_params()

            # Mock the return values
            mock_router_get_rotation_schedules.return_value.schedules = [
                MagicMock(
                    recordUid=utils.base64_url_decode('record_uid'),
                    controllerUid=utils.base64_url_decode('controller_uid'),
                    configurationUid=utils.base64_url_decode('config_uid'),
                    noSchedule=False,
                    scheduleData='RotateActionJob|daily.0.12.1'
                )
            ]

            mock_get_all_gateways.return_value = [
                MagicMock(controllerUid=utils.base64_url_decode('controller_uid'), controllerName='Controller Name')
            ]

            mock_router_get_connected_gateways.return_value.controllers = [
                MagicMock(controllerUid=utils.base64_url_decode('controller_uid'))
            ]

            mock_pam_configurations_get_all.return_value = [
                {'record_uid': 'config_uid', 'data_unencrypted': json.dumps({'title': 'Config Title', 'type': 'pamConfig'})}
            ]

            mock_pam_decrypt_configuration_data.return_value = {
                'title': 'Config Title',
                'type': 'pamConfig'
            }

            kwargs = {'is_verbose': True}
            self.command.execute(mock_params, **kwargs)

            self.assertTrue(mock_dump_report_data.called)
            self.assertTrue(mock_router_get_rotation_schedules.called)
            self.assertTrue(mock_get_all_gateways.called)
            self.assertTrue(mock_router_get_connected_gateways.called)
            self.assertTrue(mock_pam_configurations_get_all.called)

        @patch('keepercommander.commands.discoveryrotation.router_get_rotation_schedules')
        @patch('keepercommander.commands.discoveryrotation.router_get_connected_gateways')
        @patch('keepercommander.commands.discoveryrotation.pam_configurations_get_all')
        @patch('keepercommander.commands.discoveryrotation.gateway_helper.get_all_gateways')
        @patch('keepercommander.commands.discoveryrotation.pam_decrypt_configuration_data')
        @patch('keepercommander.commands.discoveryrotation.dump_report_data')
        def test_execute_with_no_schedules(self, mock_dump_report_data, mock_pam_decrypt_configuration_data, mock_get_all_gateways,
                                           mock_pam_configurations_get_all, mock_router_get_connected_gateways, mock_router_get_rotation_schedules):
            mock_params = create_mock_params()

            # Mock the return values
            mock_router_get_rotation_schedules.return_value.schedules = []

            mock_get_all_gateways.return_value = []

            mock_router_get_connected_gateways.return_value.controllers = []

            mock_pam_configurations_get_all.return_value = []

            mock_pam_decrypt_configuration_data.return_value = {}

            kwargs = {'is_verbose': True}
            self.command.execute(mock_params, **kwargs)

            self.assertTrue(mock_dump_report_data.called)
            self.assertTrue(mock_router_get_rotation_schedules.called)
            self.assertTrue(mock_get_all_gateways.called)
            self.assertTrue(mock_router_get_connected_gateways.called)
            self.assertTrue(mock_pam_configurations_get_all.called)


    class TestPAMGatewayListCommand(unittest.TestCase):

        def setUp(self):
            self.command = PAMGatewayListCommand()
            self.parser = self.command.get_parser()

        def test_parser(self):
            args = self.parser.parse_args(['--verbose', '--force'])
            self.assertTrue(args.is_verbose)
            self.assertTrue(args.is_force)

        @patch('keepercommander.commands.discoveryrotation.router_get_connected_gateways')
        @patch('keepercommander.commands.discoveryrotation.router_helper.get_router_url')
        @patch('keepercommander.commands.discoveryrotation.gateway_helper.get_all_gateways')
        @patch('keepercommander.commands.discoveryrotation.KSMCommand.get_app_record')
        @patch('keepercommander.commands.discoveryrotation.dump_report_data')
        def test_execute(self, mock_dump_report_data, mock_get_app_record, mock_get_all_gateways,
                         mock_get_router_url, mock_router_get_connected_gateways):
            mock_params = create_mock_params()

            # Mock the return values
            mock_router_get_connected_gateways.return_value.controllers = [
                MagicMock(controllerUid=utils.base64_url_decode('controller_uid'))
            ]

            mock_get_all_gateways.return_value = [
                MagicMock(
                    applicationUid=utils.base64_url_decode('app_uid'),
                    controllerUid=utils.base64_url_decode('controller_uid'),
                    controllerName='Controller Name',
                    deviceName='Device Name',
                    deviceToken='Device Token',
                    created=int(datetime.now().timestamp() * 1000),
                    lastModified=int(datetime.now().timestamp() * 1000),
                    nodeId='Node ID'
                )
            ]

            mock_get_app_record.return_value = {
                'data_unencrypted': json.dumps({'title': 'App Title'})
            }

            kwargs = {'is_force': True, 'is_verbose': True}
            self.command.execute(mock_params, **kwargs)

            self.assertTrue(mock_dump_report_data.called)
            self.assertTrue(mock_router_get_connected_gateways.called)
            self.assertTrue(mock_get_all_gateways.called)
            self.assertTrue(mock_get_router_url.called)
            self.assertTrue(mock_get_app_record.called)

        @patch('keepercommander.commands.discoveryrotation.router_get_connected_gateways')
        @patch('keepercommander.commands.discoveryrotation.router_helper.get_router_url')
        @patch('keepercommander.commands.discoveryrotation.gateway_helper.get_all_gateways')
        @patch('keepercommander.commands.discoveryrotation.dump_report_data')
        def test_execute_router_down(self, mock_dump_report_data, mock_get_all_gateways,
                                     mock_get_router_url, mock_router_get_connected_gateways):
            mock_params = create_mock_params()

            # Simulate a connection error
            mock_router_get_connected_gateways.side_effect = requests.exceptions.ConnectionError

            mock_get_all_gateways.return_value = [
                MagicMock(
                    applicationUid=utils.base64_url_decode('app_uid'),
                    controllerUid=utils.base64_url_decode('controller_uid'),
                    controllerName='Controller Name',
                    deviceName='Device Name',
                    deviceToken='Device Token',
                    created=int(datetime.now().timestamp() * 1000),
                    lastModified=int(datetime.now().timestamp() * 1000),
                    nodeId='Node ID'
                )
            ]

            kwargs = {'is_force': True, 'is_verbose': True}
            self.command.execute(mock_params, **kwargs)

            self.assertTrue(mock_dump_report_data.called)
            self.assertTrue(mock_router_get_connected_gateways.called)
            self.assertTrue(mock_get_all_gateways.called)
            self.assertTrue(mock_get_router_url.called)

        @patch('keepercommander.commands.discoveryrotation.router_get_connected_gateways')
        @patch('keepercommander.commands.discoveryrotation.router_helper.get_router_url')
        @patch('keepercommander.commands.discoveryrotation.gateway_helper.get_all_gateways')
        def test_execute_no_gateways(self, mock_get_all_gateways,
                                     mock_get_router_url, mock_router_get_connected_gateways):
            mock_params = create_mock_params()

            mock_router_get_connected_gateways.return_value.controllers = []

            mock_get_all_gateways.return_value = []

            kwargs = {'is_force': True, 'is_verbose': True}
            self.command.execute(mock_params, **kwargs)

            self.assertTrue(mock_router_get_connected_gateways.called)
            self.assertTrue(mock_get_all_gateways.called)
            self.assertTrue(mock_get_router_url.called)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: NotificationCenter.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import GraphSync_pb2 as GraphSync__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18NotificationCenter.proto\x12\x12NotificationCenter\x1a\x0fGraphSync.proto\".\n\rEncryptedData\x12\x0f\n\x07version\x18\x01 \x01(\x05\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\"\xe2\x02\n\x0cNotification\x12\x32\n\x04type\x18\x01 \x01(\x0e\x32$.NotificationCenter.NotificationType\x12>\n\x08\x63\x61tegory\x18\x02 \x01(\x0e\x32(.NotificationCenter.NotificationCategoryB\x02\x18\x01\x12\'\n\x06sender\x18\x03 \x01(\x0b\x32\x17.GraphSync.GraphSyncRef\x12\x16\n\x0esenderFullName\x18\x04 \x01(\t\x12\x38\n\rencryptedData\x18\x05 \x01(\x0b\x32!.NotificationCenter.EncryptedData\x12%\n\x04refs\x18\x06 \x03(\x0b\x32\x17.GraphSync.GraphSyncRef\x12<\n\ncategories\x18\x07 \x03(\x0e\x32(.NotificationCenter.NotificationCategory\"\x97\x01\n\x14NotificationReadMark\x12\x0b\n\x03uid\x18\x01 \x01(\x0c\x12\x1c\n\x14notification_edge_id\x18\x02 \x01(\x03\x12\x14\n\x0cmark_edge_id\x18\x03 \x01(\x03\x12>\n\nreadStatus\x18\x04 \x01(\x0e\x32*.NotificationCenter.NotificationReadStatus\"\x8d\x02\n\x13NotificationContent\x12\x38\n\x0cnotification\x18\x01 \x01(\x0b\x32 .NotificationCenter.NotificationH\x00\x12@\n\nreadStatus\x18\x02 \x01(\x0e\x32*.NotificationCenter.NotificationReadStatusH\x00\x12H\n\x0e\x61pprovalStatus\x18\x03 \x01(\x0e\x32..NotificationCenter.NotificationApprovalStatusH\x00\x12\x15\n\rclientTypeIDs\x18\x04 \x03(\x05\x12\x11\n\tdeviceIDs\x18\x05 \x03(\x03\x42\x06\n\x04type\"o\n\x13NotificationWrapper\x12\x0b\n\x03uid\x18\x01 \x01(\x0c\x12\x38\n\x07\x63ontent\x18\x02 \x01(\x0b\x32\'.NotificationCenter.NotificationContent\x12\x11\n\ttimestamp\x18\x03 \x01(\x03\"m\n\x10NotificationSync\x12\x35\n\x04\x64\x61ta\x18\x01 \x03(\x0b\x32\'.NotificationCenter.NotificationWrapper\x12\x11\n\tsyncPoint\x18\x02 \x01(\x03\x12\x0f\n\x07hasMore\x18\x03 \x01(\x08\"g\n\x10ReadStatusUpdate\x12\x17\n\x0fnotificationUid\x18\x01 \x01(\x0c\x12:\n\x06status\x18\x02 \x01(\x0e\x32*.NotificationCenter.NotificationReadStatus\"o\n\x14\x41pprovalStatusUpdate\x12\x17\n\x0fnotificationUid\x18\x01 \x01(\x0c\x12>\n\x06status\x18\x02 \x01(\x0e\x32..NotificationCenter.NotificationApprovalStatus\"^\n\x1cProcessMarkReadEventsRequest\x12>\n\x10readStatusUpdate\x18\x01 \x03(\x0b\x32$.NotificationCenter.ReadStatusUpdate\"\xa8\x01\n\x17NotificationSendRequest\x12+\n\nrecipients\x18\x01 \x03(\x0b\x32\x17.GraphSync.GraphSyncRef\x12\x36\n\x0cnotification\x18\x02 \x01(\x0b\x32 .NotificationCenter.Notification\x12\x15\n\rclientTypeIDs\x18\x03 \x03(\x05\x12\x11\n\tdeviceIDs\x18\x04 \x03(\x03\"^\n\x18NotificationsSendRequest\x12\x42\n\rnotifications\x18\x01 \x03(\x0b\x32+.NotificationCenter.NotificationSendRequest\",\n\x17NotificationSyncRequest\x12\x11\n\tsyncPoint\x18\x01 \x01(\x03*\x9f\x01\n\x14NotificationCategory\x12\x12\n\x0eNC_UNSPECIFIED\x10\x00\x12\x0e\n\nNC_ACCOUNT\x10\x01\x12\x0e\n\nNC_SHARING\x10\x02\x12\x11\n\rNC_ENTERPRISE\x10\x03\x12\x0f\n\x0bNC_SECURITY\x10\x04\x12\x0e\n\nNC_REQUEST\x10\x05\x12\r\n\tNC_SYSTEM\x10\x06\x12\x10\n\x0cNC_PROMOTION\x10\x07*\xc2\x02\n\x10NotificationType\x12\x12\n\x0eNT_UNSPECIFIED\x10\x00\x12\x0c\n\x08NT_ALERT\x10\x01\x12\x16\n\x12NT_DEVICE_APPROVAL\x10\x02\x12\x1a\n\x16NT_MASTER_PASS_UPDATED\x10\x03\x12\x15\n\x11NT_SHARE_APPROVAL\x10\x04\x12\x1e\n\x1aNT_SHARE_APPROVAL_APPROVED\x10\x05\x12\r\n\tNT_SHARED\x10\x06\x12\x12\n\x0eNT_TRANSFERRED\x10\x07\x12\x1c\n\x18NT_LICENSE_LIMIT_REACHED\x10\x08\x12\x17\n\x13NT_APPROVAL_REQUEST\x10\t\x12\x18\n\x14NT_APPROVED_RESPONSE\x10\n\x12\x16\n\x12NT_DENIED_RESPONSE\x10\x0b\x12\x15\n\x11NT_2FA_CONFIGURED\x10\x0c*Y\n\x16NotificationReadStatus\x12\x13\n\x0fNRS_UNSPECIFIED\x10\x00\x12\x0c\n\x08NRS_LAST\x10\x01\x12\x0c\n\x08NRS_READ\x10\x02\x12\x0e\n\nNRS_UNREAD\x10\x03*\x86\x01\n\x1aNotificationApprovalStatus\x12\x13\n\x0fNAS_UNSPECIFIED\x10\x00\x12\x10\n\x0cNAS_APPROVED\x10\x01\x12\x0e\n\nNAS_DENIED\x10\x02\x12\x1c\n\x18NAS_LOST_APPROVAL_RIGHTS\x10\x03\x12\x13\n\x0fNAS_LOST_ACCESS\x10\x04\x42.\n\x18\x63om.keepersecurity.protoB\x12NotificationCenterb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'NotificationCenter_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030com.keepersecurity.protoB\022NotificationCenter'
  _globals['_NOTIFICATION'].fields_by_name['category']._options = None
  _globals['_NOTIFICATION'].fields_by_name['category']._serialized_options = b'\030\001'
  _globals['_NOTIFICATIONCATEGORY']._serialized_start=1748
  _globals['_NOTIFICATIONCATEGORY']._serialized_end=1907
  _globals['_NOTIFICATIONTYPE']._serialized_start=1910
  _globals['_NOTIFICATIONTYPE']._serialized_end=2232
  _globals['_NOTIFICATIONREADSTATUS']._serialized_start=2234
  _globals['_NOTIFICATIONREADSTATUS']._serialized_end=2323
  _globals['_NOTIFICATIONAPPROVALSTATUS']._serialized_start=2326
  _globals['_NOTIFICATIONAPPROVALSTATUS']._serialized_end=2460
  _globals['_ENCRYPTEDDATA']._serialized_start=65
  _globals['_ENCRYPTEDDATA']._serialized_end=111
  _globals['_NOTIFICATION']._serialized_start=114
  _globals['_NOTIFICATION']._serialized_end=468
  _globals['_NOTIFICATIONREADMARK']._serialized_start=471
  _globals['_NOTIFICATIONREADMARK']._serialized_end=622
  _globals['_NOTIFICATIONCONTENT']._serialized_start=625
  _globals['_NOTIFICATIONCONTENT']._serialized_end=894
  _globals['_NOTIFICATIONWRAPPER']._serialized_start=896
  _globals['_NOTIFICATIONWRAPPER']._serialized_end=1007
  _globals['_NOTIFICATIONSYNC']._serialized_start=1009
  _globals['_NOTIFICATIONSYNC']._serialized_end=1118
  _globals['_READSTATUSUPDATE']._serialized_start=1120
  _globals['_READSTATUSUPDATE']._serialized_end=1223
  _globals['_APPROVALSTATUSUPDATE']._serialized_start=1225
  _globals['_APPROVALSTATUSUPDATE']._serialized_end=1336
  _globals['_PROCESSMARKREADEVENTSREQUEST']._serialized_start=1338
  _globals['_PROCESSMARKREADEVENTSREQUEST']._serialized_end=1432
  _globals['_NOTIFICATIONSENDREQUEST']._serialized_start=1435
  _globals['_NOTIFICATIONSENDREQUEST']._serialized_end=1603
  _globals['_NOTIFICATIONSSENDREQUEST']._serialized_start=1605
  _globals['_NOTIFICATIONSSENDREQUEST']._serialized_end=1699
  _globals['_NOTIFICATIONSYNCREQUEST']._serialized_start=1701
  _globals['_NOTIFICATIONSYNCREQUEST']._serialized_end=1745
# @@protoc_insertion_point(module_scope)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import io
import secrets

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import CBC, GCM
from cryptography.hazmat.primitives.hashes import Hash, SHA256, SHA512
from cryptography.hazmat.primitives.hmac import HMAC
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.primitives.padding import PKCS7

_CRYPTO_BACKEND = default_backend()
_CURVE = ec.SECP256R1()


def pad_data(data):    # type: (bytes) -> bytes
    padder = PKCS7(16*8).padder()
    return padder.update(data) + padder.finalize()


def unpad_data(data):     # type: (bytes) -> bytes
    unpadder = PKCS7(16*8).unpadder()
    return unpadder.update(data) + unpadder.finalize()


def get_random_bytes(length):
    return secrets.token_bytes(length)


def generate_rsa_key():
    private_key = rsa.generate_private_key(key_size=2048, public_exponent=0x10001, backend=_CRYPTO_BACKEND)
    return private_key, private_key.public_key()


def load_rsa_private_key(der_private_key, password=None):
    return serialization.load_der_private_key(der_private_key, password, _CRYPTO_BACKEND)


def load_rsa_public_key(der_public_key):
    return serialization.load_der_public_key(der_public_key, _CRYPTO_BACKEND)


def unload_rsa_private_key(private_key):
    return private_key.private_bytes(encoding=serialization.Encoding.DER,
                                     format=serialization.PrivateFormat.TraditionalOpenSSL,
                                     encryption_algorithm=serialization.NoEncryption())


def unload_rsa_public_key(public_key):
    return public_key.public_bytes(encoding=serialization.Encoding.DER,
                                   format=serialization.PublicFormat.PKCS1)


def generate_ec_key():
    private_key = ec.generate_private_key(curve=_CURVE, backend=_CRYPTO_BACKEND)
    return private_key, private_key.public_key()


def load_ec_private_key(private_key):
    private_value = int.from_bytes(private_key, byteorder='big', signed=False)
    return ec.derive_private_key(private_value, _CURVE, _CRYPTO_BACKEND)


def load_ec_public_key(public_key):
    return ec.EllipticCurvePublicKey.from_encoded_point(_CURVE, public_key)


def unload_ec_private_key(private_key):
    private_value = private_key.private_numbers().private_value
    return private_value.to_bytes(length=32, byteorder='big', signed=False)


def unload_ec_public_key(public_key):
    return public_key.public_bytes(encoding=serialization.Encoding.X962,
                                   format=serialization.PublicFormat.UncompressedPoint)


def encrypt_aes_v1(data, key, iv=None, use_padding=True):
    iv = iv or get_random_bytes(16)
    cipher = Cipher(AES(key), CBC(iv), backend=_CRYPTO_BACKEND)
    encrypter = cipher.encryptor()
    encrypted_data = encrypter.update(pad_data(data) if use_padding else data) + encrypter.finalize()
    return iv + encrypted_data


def decrypt_aes_v1(data, key, use_padding=True):
    iv = data[:16]
    cipher = Cipher(AES(key), CBC(iv), backend=_CRYPTO_BACKEND)
    decrypter = cipher.decryptor()
    decrypted_data = decrypter.update(data[16:]) + decrypter.finalize()
    return unpad_data(decrypted_data) if use_padding else decrypted_data


def encrypt_aes_v2(data, key, nonce=None):
    nonce = nonce or get_random_bytes(12)
    cipher = Cipher(AES(key), GCM(nonce), backend=_CRYPTO_BACKEND)
    encrypter = cipher.encryptor()
    encrypted_data = encrypter.update(data) + encrypter.finalize()
    return nonce + encrypted_data + encrypter.tag


def decrypt_aes_v2(data, key):
    nonce = data[:12]
    cipher = Cipher(AES(key), GCM(nonce), backend=_CRYPTO_BACKEND)
    decrypter = cipher.decryptor()
    decrypted_data = decrypter.update(data[12:-16]) + decrypter.finalize_with_tag(data[-16:])
    return decrypted_data


def encrypt_rsa(data, rsa_key):
    return rsa_key.encrypt(data, PKCS1v15())


def decrypt_rsa(data, rsa_key, apply_padding=True):
    size_diff = (rsa_key.key_size + 7 >> 3) - len(data)
    if apply_padding and size_diff > 0:
        pad_bytes = bytes(size_diff)
        pad_bytearray = bytearray([b for b in pad_bytes])
        data_bytearray = bytearray([b for b in data])
        data = bytes(pad_bytearray + data_bytearray)
    return rsa_key.decrypt(data, PKCS1v15())


def encrypt_ec(data, ec_public_key):
    e_private_key, e_public_key = generate_ec_key()
    shared_secret = e_private_key.exchange(ec.ECDH(), ec_public_key)
    digest = Hash(SHA256(), backend=_CRYPTO_BACKEND)
    digest.update(shared_secret)
    encryption_key = digest.finalize()
    return unload_ec_public_key(e_public_key) + encrypt_aes_v2(data, encryption_key)


def decrypt_ec(data, ec_private_key):
    ephemeral_public_key = load_ec_public_key(data[:65])
    shared_secret = ec_private_key.exchange(ec.ECDH(), ephemeral_public_key)
    digest = Hash(SHA256(), backend=_CRYPTO_BACKEND)
    digest.update(shared_secret)
    encryption_key = digest.finalize()
    return decrypt_aes_v2(data[65:], encryption_key)


def derive_key_v1(password, salt, iterations):
    kdf = PBKDF2HMAC(algorithm=SHA256(), length=32, salt=salt, iterations=iterations, backend=_CRYPTO_BACKEND)
    return kdf.derive(password.encode('utf-8'))


def derive_keyhash_v1(password, salt, iterations):
    derived_key = derive_key_v1(password, salt, iterations)
    hf = Hash(SHA256(), backend=_CRYPTO_BACKEND)
    hf.update(derived_key)
    return hf.finalize()


def derive_keyhash_v2(domain, password, salt, iterations):
    kdf = PBKDF2HMAC(algorithm=SHA512(), length=64, salt=salt, iterations=iterations, backend=_CRYPTO_BACKEND)
    derived_key = kdf.derive((domain+password).encode('utf-8'))
    hf = HMAC(derived_key, SHA256(), backend=_CRYPTO_BACKEND)
    hf.update(domain.encode('utf-8'))
    return hf.finalize()


def hmac_sha512(key, data):
    hf = HMAC(key, SHA512(), backend=_CRYPTO_BACKEND)
    hf.update(data)
    return hf.finalize()


def generate_hkdf_key(info, phrase):
    hf = HKDF(SHA512(), length=32, salt=b'', info=info.encode('utf-8'), backend=_CRYPTO_BACKEND)
    return hf.derive(phrase.encode('utf-8'))


class _StreamCrypter(io.RawIOBase):
    def __init__(self):
        super().__init__()
        self.key = b''
        self.is_gcm = False
        self.is_encrypt = False
        self.bytes_read = 0
        self._base_stream = None
        self.crypter = None
        self.is_eof = False
        self.in_buffer = None
        self.out_buffer = None
        self.in_buffer_pos = 0
        self.out_buffer_pos = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.in_buffer = None
        self.out_buffer = None
        self.crypter = None
        if self._base_stream:
            if hasattr(self._base_stream, '__exit__'):
                self._base_stream.__exit__(exc_type, exc_val, exc_tb)
            elif hasattr(self._base_stream, 'close'):
                self._base_stream.close()
            self._base_stream = None

    def set_stream(self, stream, for_encrypt):
        self.in_buffer = memoryview(bytearray(10240))
        self.in_buffer_pos = 0
        self.out_buffer = memoryview(bytearray(10240))
        self.out_buffer_pos = 0
        self.is_encrypt = for_encrypt
        self.bytes_read = 0
        if stream:
            self.is_eof = False
            if self.is_gcm:
                if self.is_encrypt:
                    nonce = get_random_bytes(12)
                    self.out_buffer[self.out_buffer_pos:self.out_buffer_pos + len(nonce)] = nonce
                    self.out_buffer_pos += len(nonce)
                else:
                    nonce = stream.read(12)
                    self.bytes_read += len(nonce)
                cipher = Cipher(AES(self.key), GCM(nonce), backend=_CRYPTO_BACKEND)
            else:
                if self.is_encrypt:
                    iv = get_random_bytes(16)
                    self.out_buffer[self.out_buffer_pos:self.out_buffer_pos + len(iv)] = iv
                    self.out_buffer_pos += len(iv)
                else:
                    iv = stream.read(16)
                    self.bytes_read += len(iv)
                cipher = Cipher(AES(self.key), CBC(iv), backend=_CRYPTO_BACKEND)
            self.crypter = cipher.encryptor() if self.is_encrypt else cipher.decryptor()
            self._base_stream = stream
        else:
            self.is_eof = True
        return self

    def closed(self):
        return self._base_stream is not None

    def close(self):
        self.__exit__(None, None, None)

    def readinto(self, buffer):
        buffer_len = 0
        mv = memoryview(buffer)
        while buffer_len < len(buffer):
            if self.out_buffer_pos == 0:
                if self._base_stream and not self.is_eof:
                    view = self.in_buffer[self.in_buffer_pos:len(self.in_buffer) - self.in_buffer_pos]
                    bytes_read = self._base_stream.readinto(view)
                    self.bytes_read += bytes_read
                    self.is_eof = bytes_read == 0
                    self.in_buffer_pos += bytes_read
                to_crypt = max(self.in_buffer_pos - 16, 0)
                if not self.is_gcm:
                    rem = to_crypt % 16
                    if rem != 0:
                        to_crypt -= rem
                bytes_left = len(self.out_buffer) - self.out_buffer_pos
                to_crypt = min(to_crypt, bytes_left - (16 if self.is_encrypt else 0))
                crypted = None
                if to_crypt > 0:
                    crypted = self.crypter.update(self.in_buffer[0:to_crypt])
                    if self.in_buffer_pos > to_crypt:
                        self.in_buffer[0:self.in_buffer_pos - to_crypt] = self.in_buffer[to_crypt: self.in_buffer_pos]
                        self.in_buffer_pos -= to_crypt
                    else:
                        self.in_buffer_pos = 0
                else:
                    if self.is_encrypt:
                        if self.in_buffer_pos > 0:
                            tail = bytes(self.in_buffer[0:self.in_buffer_pos])
                            self.in_buffer_pos = 0
                            if not self.is_gcm:
                                tail = pad_data(tail)
                            crypted = self.crypter.update(tail)
                            crypted += self.crypter.finalize()
                            if self.is_gcm:
                                crypted += self.crypter.tag
                    else:
                        if self.in_buffer_pos == 16:
                            if self.is_gcm:
                                tag = bytes(self.in_buffer[0: self.in_buffer_pos])
                                crypted = self.crypter.finalize_with_tag(tag)
                            else:
                                crypted = self.crypter.update(self.in_buffer[0: self.in_buffer_pos]) + self.crypter.finalize()
                                crypted = unpad_data(crypted)
                            self.in_buffer_pos = 0
                if crypted:
                    self.out_buffer[self.out_buffer_pos: self.out_buffer_pos + len(crypted)] = crypted
                    self.out_buffer_pos += len(crypted)

            if self.out_buffer_pos > 0:
                b_len = min(len(buffer) - buffer_len, self.out_buffer_pos)
                mv[buffer_len:buffer_len+b_len] = self.out_buffer[0:b_len]
                if self.out_buffer_pos > b_len:
                    self.out_buffer[0:self.out_buffer_pos - b_len] = self.out_buffer[b_len:self.out_buffer_pos]
                    self.out_buffer_pos -= b_len
                else:
                    self.out_buffer_pos = 0
                buffer_len += b_len
            else:
                break

        return buffer_len


class StreamCrypter(_StreamCrypter):
    __doc__ = _StreamCrypter.__doc__



from __future__ import annotations
import logging
from .types import EdgeType
from .exceptions import DAGContentException
import json
from typing import Optional, Union, Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from .vertex import DAGVertex
    Content = Union[str, bytes, dict]
    QueryValue = Union[list, dict, str, float, int, bool]
    import pydantic
    from pydantic import BaseModel


class DAGEdge:
    def __init__(self, vertex: DAGVertex, edge_type: EdgeType, head_uid: str, version: int = 0,
                 content: Optional[Any] = None, path: Optional[str] = None,
                 modified: bool = True, block_content_auto_save: bool = False, from_load: bool = False,
                 needs_encryption: bool = False):
        """
        Create an instance of DAGEdge.

        A primary key of the edge the vertex UID, the head UID, and edge_type.

        :param vertex: The DAGVertex instance that owns these edges.
        :param edge_type: The enumeration EdgeType. Indicate the type of the edge.
        :param head_uid: The vertex uid that has this edge's vertex. The vertex uid that the edge arrow points at.
        :param version: Version of this edge.
        :param content: The content of this edge.
        :param path: Short tag about this edge. Do
        :param modified:
        :param block_content_auto_save:
        :param from_load: Is this being called from the load() method?
        :param needs_encryption: Flag to indicate if the content needs to be encrypted.
        :return: An instance of DAGEdge
        """

        # This is the vertex that owns this edge.
        self.vertex = vertex
        self.edge_type = edge_type
        self.head_uid = head_uid

        # Flag to indicate if the edge has been modified. Used to determine if the edge should be part of saved data.
        # Set this before setting the content, else setting the content will cause an auto save.
        self._modified = None
        self.modified = modified

        # Block auto save in the content setter.
        # When creating an edge, don't save until the edge is added to the edge list.
        self.block_content_auto_save = block_content_auto_save

        # Does this edge's content need encryption?
        self.needs_encryption = needs_encryption

        # If the content is being populated from a the load() method, and the edge type is a KEY or DATA, then the
        # content will be encrypted (str).
        # We want to keep a str, unless KEYs are decrypted.

        # If the edge data need encryption, is _content, currently encrypted.
        self.encrypted = from_load is True and edge_type in [EdgeType.KEY, EdgeType.DATA]

        # If the content could not be decrypted, set
        self.corrupt = False

        self._content = None  # type: Optional[Any]
        self.content = content
        self.path = path

        self.version = version

        # If a higher version edge exists, this will be False.
        # If True, this is the highest edge.
        self.active = True

    def __str__(self) -> str:
        return f"<Edge type {self.vertex.dag.__class__.EDGE_LABEL.get(self.edge_type)}, head {self.head_uid}, "\
               f"tail {self.vertex.uid}, path {self.path}, version {self.version}>"

    def debug(self, msg, level=0):
        self.vertex.dag.debug(msg, level=level)

    @property
    def modified(self):
        return self._modified

    @modified.setter
    def modified(self, value):
        if value is True:
            self.debug(f"vertex {self.vertex.uid}, type {self.vertex.dag.__class__.EDGE_LABEL.get(self.edge_type)}, "
                       f"head {self.head_uid} has been modified", level=5)
        else:
            self.debug(f"vertex {self.vertex.uid}, type {self.vertex.dag.__class__.EDGE_LABEL.get(self.edge_type)}, "
                       f"head {self.head_uid} had modified RESET", level=5)
        self._modified = value

    @property
    def content(self) -> Optional[Union[str, bytes]]:
        """
        Get the content of the edge.

        If the content is a str, then the content is encrypted.
        """
        return self._content

    @property
    def content_as_dict(self) -> Optional[dict]:
        """
        Get the content from the DATA edge as a dictionary.
        :return: Content as a dictionary.
        """
        content = self._content
        if content is not None:
            try:
                content = json.loads(content)
            except Exception as err:
                raise DAGContentException(f"Cannot decode JSON. Is the content a dictionary? : {err}")
        return content

    @property
    def content_as_str(self) -> Optional[str]:
        """
        Get the content from the DATA edge as string
        :return:
        """
        content = self._content
        try:
            content = content.decode()
        except Exception as err:
            pass
        return content

    def content_as_object(self, meta_class: pydantic._internal._model_construction.ModelMetaclass) -> (
            Optional)[BaseModel]:
        """
        Get the content as a pydantic based object.

        :param meta_class: The class to return
        :return:
        """
        content = self.content_as_str
        if content is not None:
            content = meta_class.model_validate_json(self.content_as_str)
        return content

    @content.setter
    def content(self, value: Any):

        """
        Set the content in the edge.

        The content should be stored as bytes.
        If the encrypted flag is set, the content will be stored as is.
        Content that is a str type is encrypted data (A Base64, AES encrypted bytes, str)
        """

        self.debug(f"vertex {self.vertex.uid}, type {self.vertex.dag.__class__.EDGE_LABEL.get(self.edge_type)}, "
                   f"head {self.head_uid} setting content", level=2)

        # If the data is encrypted, set it.
        # Don't try to make it bytes.
        # Also don't set the modified flag to True.
        if self.encrypted is True:
            self.debug("  content is encrypted.", level=3)
            self._content = value
            return

        if self._content is not None:
            raise DAGContentException("Cannot update existing content. Use add_data() to change the content.")

        if isinstance(value, dict) is True:
            value = json.dumps(value)

        # Is this a Pydantic based class?
        if hasattr(value, "model_dump_json") is True:
            value = value.model_dump_json()

        if isinstance(value, str) is True:
            value = value.encode()

        self._content = value

    def delete(self):
        """
        Delete the edge.

        Deleting an edge does not remove the existing edge.
        It will create another edge with the same tail and head, but will be type DELETION.
        """

        # If already inactive, return
        if self.active is False:
            return

        version, _ = self.vertex.get_highest_edge_version(head_uid=self.head_uid)

        # Flag all other edges as inactive.
        for edge in self.vertex.edges:
            edge.active = False

        self.vertex.edges.append(
            DAGEdge(
                vertex=self.vertex,
                edge_type=EdgeType.DELETION,
                head_uid=self.head_uid,
                version=version + 1
            )
        )

        # Perform the DELETION edges save in one batch.
        # Get the current allowed auto save state, and disable auto save.
        current_allow_auto_save = self.vertex.dag.allow_auto_save
        self.vertex.dag.allow_auto_save = False

        if self.vertex.belongs_to_a_vertex is False:
            self.vertex.delete(ignore_vertex=self.vertex)

        self.vertex.dag.allow_auto_save = current_allow_auto_save
        self.vertex.dag.do_auto_save()

    @property
    def is_deleted(self) -> bool:
        """
        Does this edge have a DELETION edge that has the same head?

        This should be used to check in a non-DELETION edge type has a matching DELETION edge.
        :return:
        """

        # We shouldn't be checking the DELETION edge if it deleted.
        # Throw some info message to make sure the coder knows their code might be something foolish.
        if self.edge_type == EdgeType.DELETION:
            logging.info(f"The edge is_deleted() just check if the DELETION edge is DELETION "
                         f"for vertex {self.vertex.uid}, head UID {self.head_uid}. Returned True, but code should "
                         "not be checking this edge.")
            return True

        # Check the other edges for this vertex for an active DELETION-edge type.
        for edge in self.vertex.edges:
            if edge.edge_type == EdgeType.DELETION and edge.head_uid == self.head_uid and edge.active is True:
                return True

        return False



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#
import abc
import argparse
import base64
import datetime
import io
import json
import logging
import os
from typing import Iterable, Union, Optional, Dict, List, Set, Any, Callable
from urllib.parse import urlparse, urlunparse, parse_qsl, urlencode

import requests

from .base import user_choice, dump_report_data, report_output_parser, field_to_title, GroupCommand
from .enterprise import TeamApproveCommand, EnterpriseCommand
from .. import api, utils, vault, attachment, vault_extensions
from ..display import bcolors
from ..error import CommandError
from ..params import KeeperParams

scim_list_parser = argparse.ArgumentParser(prog='scim list', parents=[report_output_parser],
                                           description='Display a list of available SCIM endpoints.')

scim_view_parser = argparse.ArgumentParser(prog='scim view', description='Display a SCIM endpoint details.')
scim_view_parser.add_argument('target', help='SCIM ID')
scim_view_parser.add_argument('--format', dest='format', action='store', choices=['table', 'json'], default='table',
                              help='output format.')
scim_view_parser.add_argument('--output', dest='output', action='store',
                              help='output file name. (ignored for table format)')

scim_create_parser = argparse.ArgumentParser(prog='scim create', description='Create SCIM endpoint.')
scim_create_parser.add_argument('--node', dest='node', required=True, help='Node Name or ID.')
scim_create_parser.add_argument(
    '--prefix', dest='prefix', action='store',
    help='Role Prefix. SCIM groups staring with prefix will be imported to Keeper as Roles.')
scim_create_parser.add_argument(
    '--unique-groups', dest='unique_groups', action='store', choices=['on', 'off'], help='Unique Groups.')

scim_edit_parser = argparse.ArgumentParser(prog='scim edit', description='Edit SCIM endpoint.')
scim_edit_parser.add_argument('target', help='SCIM ID')
scim_edit_parser.add_argument(
    '--prefix', dest='prefix', action='store',
    help='Role Prefix. SCIM groups staring with prefix will be imported to Keeper as Roles.')
scim_edit_parser.add_argument(
    '--unique-groups', dest='unique_groups', action='store', choices=['on', 'off'], help='Unique Groups.')

scim_delete_parser = argparse.ArgumentParser(prog='scim delete', description='Delete SCIM endpoint.')
scim_delete_parser.add_argument('target', help='SCIM ID')
scim_delete_parser.add_argument('--force', '-f', dest='force', action='store_true', help='Delete with no confirmation')

scim_push_parser = argparse.ArgumentParser(prog='scim push', description='Push data to SCIM endpoint.')
scim_push_parser.add_argument('--dry-run', dest='dry_run', action='store_true',
                              help='display SCIM requests without posing them')
scim_push_parser.add_argument('--source', dest='source', action='store', choices=['google', 'ad', 'record'],
                              default='google', help='Source of SCIM data')
scim_push_parser.add_argument('--record', '-r', dest='record', action='store',
                              help='Record UID with SCIM configuration')
scim_push_parser.add_argument('--auto-approve', dest='auto_approve', action='store', choices=['on', 'off'],
                              default='on', help='Auto approve SCIM teams')
scim_push_parser.add_argument('target', help='SCIM ID')


def register_commands(commands):
    commands['scim'] = ScimCommand()


def register_command_info(aliases, command_info):
    command_info['scim'] = 'Manage SCIM endpoints'


class ScimCommand(GroupCommand):
    def __init__(self):
        super(ScimCommand, self).__init__()
        self.register_command('list', ScimListCommand())
        self.register_command('view', ScimViewCommand())
        self.register_command('create', ScimCreateCommand())
        self.register_command('edit', ScimEditCommand())
        self.register_command('delete', ScimDeleteCommand())
        self.register_command('push', ScimPushCommand())
        self.default_verb = 'list'


class ScimListCommand(EnterpriseCommand):
    def get_parser(self):
        return scim_list_parser

    def execute(self, params, **kwargs):
        fmt = kwargs.get('format') or ''
        headers = ['scim_id', 'node_name', 'node_id', 'prefix', 'status', 'last_synced']
        if fmt != 'json':
            headers = [field_to_title(x) for x in headers]

        table = []
        if params.enterprise and 'scims' in params.enterprise:
            for info in params.enterprise['scims']:
                node_id = info['node_id']
                last_synced = info.get('last_synced')
                if isinstance(last_synced, int):
                    dt = datetime.datetime.fromtimestamp(last_synced / 1000)
                    last_synced = dt.strftime('%c')
                else:
                    last_synced = str(last_synced or '')
                row = [info['scim_id'], self.get_node_path(params, node_id), node_id,
                       info.get('role_prefix') or '', info['status'], last_synced]
                table.append(row)
        return dump_report_data(table, headers=headers, fmt=fmt, filename=kwargs.get('output'))


class ScimCreateCommand(EnterpriseCommand):
    def get_parser(self):  # type: () -> argparse.ArgumentParser | None
        return scim_create_parser

    def execute(self, params, node=None, **kwargs):
        if not node:
            logging.warning('\"--node\" option is required for \"create\" command')
            return
        nodes = list(self.resolve_nodes(params, node))
        if len(nodes) > 1:
            logging.warning('Node name \'%s\' is not unique. Use Node ID.', node)
            return
        elif len(nodes) == 0:
            logging.warning('Node name \'%s\' is not found', node)
            return

        matched_node = nodes[0]
        if not matched_node.get('parent_id'):
            logging.warning('Root node cannot be used for SCIM endpoint')
            return
        token = base64.urlsafe_b64encode(os.urandom(32)).decode().rstrip('=')
        rq = {
            'command': 'scim_add',
            'scim_id': self.get_enterprise_id(params),
            'node_id': matched_node['node_id'],
            'token': token,
        }
        prefix = kwargs.get('prefix')
        if prefix:
            rq['prefix'] = prefix
        if kwargs.get('unique_groups', '') == 'on':
            rq['unique_groups'] = True

        api.communicate(params, rq)
        api.query_enterprise(params)
        logging.info('')
        logging.info('SCIM ID: %d', rq['scim_id'])
        logging.info('SCIM URL: %s', get_scim_url(params, matched_node['node_id']))
        logging.info('Provisioning Token: %s', token)
        logging.info('')
        return token


def find_scim(param, name):  # type: (KeeperParams, any) -> dict
    if not name:
        raise Exception('SCIM ID cannot be empty')
    if 'scims' in param.enterprise:
        try:
            scim_id = int(name)
            for scim in param.enterprise['scims']:
                if scim['scim_id'] == scim_id:
                    return scim
                if scim['node_id'] == scim_id:
                    return scim
        except:
            raise Exception('SCIM ID should be an integer: {0}'.format(name))
    raise Exception('SCIM endpoint with ID \"{0}\" not found'.format(name))


class ScimViewCommand(EnterpriseCommand):
    def get_parser(self):  # type: () -> argparse.ArgumentParser | None
        return scim_view_parser

    def execute(self, params, target=None, **kwargs):
        scim = find_scim(params, target)
        fmt = kwargs.get('format')
        node_id = scim['node_id']
        node_name = self.get_node_path(params, node_id)
        last_synced = scim.get('last_synced')
        if isinstance(last_synced, int):
            last_synced = datetime.datetime.fromtimestamp(last_synced / 1000)
        else:
            last_synced = None

        if fmt == 'json':
            j_output = {
                'scim_id': scim['scim_id'],
                'scim_url': get_scim_url(params, node_id),
                'node_id': node_id,
                'node_name': node_name,
                'status': scim['status'],
                'prefix': scim.get('role_prefix') or '',
                'unique_groups': scim.get('unique_groups', False)
            }
            if last_synced:
                j_output['last_synced'] = last_synced.strftime('%c')
            output_file = kwargs.get('output')
            if output_file:
                with open(output_file, 'w') as f:
                    json.dump(j_output, f, indent=2)
                logging.info('File name: %s', os.path.abspath(output_file))
            else:
                return json.dumps(j_output, indent=2)
        else:
            table = [
                ['SCIM ID', scim['scim_id']], ['SCIM URL', get_scim_url(params, node_id)], ['Node ID', node_id],
                ['Node Name', node_name],
                ['Status', scim['status']], ['Prefix', scim.get('role_prefix') or ''],
                ['Unique Groups', scim.get('unique_groups', False)]
            ]
            if last_synced:
                table.append(['Last Synced', last_synced.strftime('%c')])
            dump_report_data(table, ['key', 'value'], no_header=True, right_align=(0,))


class ScimEditCommand(EnterpriseCommand):
    def get_parser(self):  # type: () -> argparse.ArgumentParser | None
        return scim_edit_parser

    def execute(self, params, target=None, **kwargs):
        scim = find_scim(params, target)
        token = base64.urlsafe_b64encode(os.urandom(32)).decode().rstrip('=')
        rq = {
            'command': 'scim_update',
            'scim_id': scim['scim_id'],
            'token': token,
        }
        prefix = kwargs.get('prefix')
        if prefix:
            rq['prefix'] = prefix
        unique_groups = kwargs.get('unique_groups')
        if unique_groups:
            rq['unique_groups'] = unique_groups == 'on'

        api.communicate(params, rq)
        api.query_enterprise(params)
        logging.info('')
        logging.info('SCIM ID: %d', scim['scim_id'])
        logging.info('SCIM URL: %s', get_scim_url(params, scim['node_id']))
        logging.info('Provisioning Token: %s', token)
        logging.info('')
        return token


class ScimDeleteCommand(EnterpriseCommand):
    def get_parser(self):  # type: () -> argparse.ArgumentParser | None
        return scim_delete_parser

    def execute(self, params, target=None, **kwargs):
        scim = find_scim(params, target)
        answer = 'y' if kwargs.get('force') else \
            user_choice(bcolors.FAIL + bcolors.BOLD + '\nALERT!\n' + bcolors.ENDC +
                        'You are about to delete SCIM endpoint {0}'.format(scim['scim_id']) +
                        '\n\nDo you want to proceed with deletion?', 'yn', 'n')
        if answer.lower() != 'y':
            return

        rq = {
            'command': 'scim_delete',
            'scim_id': scim['scim_id'],
        }
        api.communicate(params, rq)
        api.query_enterprise(params)
        logging.info('SCIM endpoint \"%d\" at node \"%d\" deleted', scim['scim_id'], scim['node_id'])


def get_scim_url(params, node_id):  # type:  (KeeperParams, int) -> any
    p = urlparse(params.rest_context.server_base)
    return urlunparse((p.scheme, p.netloc, '/api/rest/scim/v2/' + str(node_id), None, None, None))


class ScimUser:
    def __init__(self):
        self.id = ''
        self.external_id = ''
        self.email = ''
        self.full_name = ''
        self.first_name = ''
        self.last_name = ''
        self.active = False
        self.groups = []

    def __str__(self):
        scim_user = {
            'id': self.id,
            'external_id': self.external_id,
            'email': self.email,
            'full_name': self.full_name,
            'first_name': self.first_name,
            'last_name': self.last_name,
            'active': self.active,
            'groups': self.groups,
        }
        return 'SCIM USER: ' + json.dumps(scim_user)


class ScimGroup:
    def __init__(self):
        self.id = ''
        self.external_id = ''
        self.name = ''

    def __str__(self):
        scim_group = {
            'id': self.id,
            'external_id': self.external_id,
            'name': self.name,
        }
        return 'SCIM GROUP: ' + json.dumps(scim_group)


class ICrmDataSource(abc.ABC):
    def __init__(self):
        self._load_errors = False     # type: bool
        self._debug_logger = ICrmDataSource.null_logger     # type: Callable[[str], None]

    @staticmethod
    def null_logger(message):
        pass

    @abc.abstractmethod
    def populate(self):        # type: () -> Iterable[Union[ScimGroup, ScimUser]]
        pass

    @property
    def load_errors(self):     # type: () -> bool
        return self._load_errors

    @property
    def debug_logger(self):    # type: () -> Callable[[str], None]
        return self._debug_logger

    @debug_logger.setter
    def debug_logger(self, value):    # type: (Callable[[str], None]) -> None
        if value is None:
            value = ICrmDataSource.null_logger
        self._debug_logger = value


class ScimPushCommand(EnterpriseCommand):
    def get_parser(self):
        return scim_push_parser

    def execute(self, params, target=None, **kwargs):
        scim = find_scim(params, target)
        dry_run = kwargs.get('dry_run') is True

        scim_url = get_scim_url(params, scim['node_id'])
        record_uid = kwargs.get('record')
        record = None  # type: Optional[vault.TypedRecord]
        if record_uid:
            if record_uid in params.record_cache:
                r = vault.KeeperRecord.load(params, record_uid)
                if isinstance(r, vault.TypedRecord):
                    record = r
                else:
                    raise CommandError('', f'Record UID "{record_uid}": invalid record type ')
            else:
                raise CommandError('', f'Record UID "{record_uid}": does not exist')
        else:
            for r in vault_extensions.find_records(params, record_version=3):
                if not isinstance(r, vault.TypedRecord):
                    continue
                field = next((x for x in r.fields if x.type == 'url'), None)
                if field:
                    url = field.get_default_value(str)
                    if url and url == scim_url:
                        record = r
                        break
            if not isinstance(record, vault.TypedRecord):
                raise CommandError('', f'Cannot find SCIM record with URL: {scim_url}')

        destructive = 0
        verbose = False
        field = next((x for x in record.fields if x.type == 'password'), None)
        if not field:
            raise CommandError('', f'"Password" field not found on record "{record.title}"')
        token = field.get_default_value(str)
        if not token:
            raise CommandError('', f'"Password" field is empty on record "{record.title}"')

        field = record.get_typed_field('text', 'Destructive')
        if field:
            sv = field.get_default_value(str)
            if sv and sv.isnumeric():
                destructive = int(sv)
        field = record.get_typed_field('text', 'Verbose')
        if field:
            sv = field.get_default_value(str)
            if sv and sv.lower() in ('true', 'yes', 'on', '1'):
                verbose = True

        def verbose_logging(message):
            if verbose:
                logging.info(message)

        keeper_users = {}  # type: Dict[str, ScimUser]
        keeper_groups = {}  # type: Dict[str, ScimGroup]
        logging.debug('SCIM Query Keeper')
        for element in ScimPushCommand.scim_keeper(scim_url, token):
            if isinstance(element, ScimUser):
                keeper_users[element.id] = element
                logging.debug(str(element))
            elif isinstance(element, ScimGroup):
                keeper_groups[element.id] = element
                logging.debug(str(element))

        source = kwargs.get('source')
        if not source:
            raise CommandError('', f'SCIM source {source} cannot be empty')

        # SCIM group
        scim_groups = []    # type: List[str]
        for cf in record.custom:
            if not cf.label:
                continue
            label = cf.label.lower().strip()
            if label.startswith('scim group'):
                group_value = cf.get_default_value(str)    # type: str
                if group_value:
                    for by_nl in group_value.split('\n'):
                        if not by_nl:
                            continue
                        scim_groups.append(by_nl.strip())
                        
        if source == 'google':
            if len(scim_groups) == 0:
                raise CommandError('', f'SCIM record "{record.title}" does not have "SCIM Group(s)" field.\n'
                                       'This field contains a group name or other group identifier with users to be imported to Keeper.')

                # Admin user
            field = record.get_typed_field('login')
            if field is None:
                raise CommandError('', f'Google SCIM record "{record.title}" does not have "login" field.\n'
                                       'Please use "login" record type to store Google SCIM configuration.')

            admin_user = field.get_default_value(str)
            if not admin_user:
                raise CommandError('', f'"login" field in Google SCIM record "{record.title}" should be populated with '
                                       f'Google Workspace administrator email')

            ad = next(attachment.prepare_attachment_download(
                params, record_uid=record.record_uid, attachment_name='credentials.json'), None)
            if not ad:
                raise CommandError('', 'Google SCIM configuration: Service account credentials are not found')

            with io.BytesIO() as mem:
                ad.download_to_stream(params, mem)
                mem.seek(0, io.SEEK_SET)
                credentials = json.load(mem)

            data_source = GoogleCrmDataSource(admin_user, credentials, scim_groups)
        elif source == 'ad':
            if len(scim_groups) == 0:
                raise CommandError('', f'SCIM record "{record.title}" does not have "SCIM Group(s)" field.\n'
                                       'This field contains a group name or other group identifier with users to be imported to Keeper.')

                # AD URL
            ad_url = ''
            field = next((x for x in record.custom if (x.label or '').lower().strip() == 'ad url'), None)
            if field:
                ad_url = field.get_default_value(str)
            if not ad_url:
                raise CommandError('', f'Active Directory SCIM record "{record.title}" does not have "AD URL" field.\n'
                                       'This field contains URL to connect to Active Directory.\n'
                                       'Format: ldap(s)://<DOMAIN_CONTROLLER_HOSTNAME_OR_IP_ADDRESS>')
            # AD User
            ad_user = ''
            field = next((x for x in record.custom if (x.label or '').lower().strip() == 'ad user'), None)
            if field:
                ad_user = field.get_default_value(str)
            if not ad_user:
                raise CommandError('', f'Active Directory SCIM record "{record.title}" does not have "AD User" field.\n'
                                       'This field contains username to connect to Active Directory.\n'
                                       'Username should be either DOMAIN\\USERNAME or user distinguished name')

            # AD Password
            ad_password = ''
            field = next((x for x in record.custom if (x.label or '').lower().strip() == 'ad password'), None)
            if field:
                ad_password = field.get_default_value(str)
            if not ad_password:
                raise CommandError('', f'Active Directory SCIM record "{record.title}" does not have "AD Password" field.\n'
                                       'This field contains AD user password.')
            data_source = AdCrmDataSource(ad_url, ad_user, ad_password, scim_groups)
        elif source == 'record':
            json_data = record.notes
            if not json_data:
                raise CommandError('', f'SCIM record "{record.title}" has empty Notes.\n'
                                       'Please store your SCIM configuration to Notes field as JSON text')
            data = json.loads(json_data)
            data_source = LocalCrmDataSource(data)
        else:
            raise CommandError('', f'SCIM source {source} is not supported')

        other_users = {}  # type: Dict[str, ScimUser]
        other_groups = {}  # type: Dict[str, ScimGroup]

        data_source.debug_logger = verbose_logging
        verbose_logging('SCIM Query External Source')
        for element in data_source.populate():
            if isinstance(element, ScimUser):
                other_users[element.id] = element
                logging.debug(str(element))
            elif isinstance(element, ScimGroup):
                other_groups[element.id] = element
                logging.debug(str(element))
        if data_source.load_errors:
            verbose_logging('Switching to the "Safe Mode" due to errors')
            destructive = -1

        self.sync_groups(scim_url, token, keeper_groups, other_groups, dry_run, destructive=destructive)
        self.sync_users(scim_url, token, keeper_users, other_users, dry_run)
        self.sync_membership(scim_url, token, keeper_groups, keeper_users, other_users, dry_run, destructive=destructive)
        api.query_enterprise(params)
        auto_approve = kwargs.get('auto_approve') or ''
        if auto_approve != 'off':
            team_approve = TeamApproveCommand()
            team_approve.execute(params)
            api.query_enterprise(params)

    @staticmethod
    def sync_groups(scim_url, token,
                    keeper_groups,
                    external_groups,
                    dry_run=False,
                    **kwargs):  # type: (str, str, Dict[str, ScimGroup], Dict[str, ScimGroup], bool, Any) -> None
        keeper_group_copy = keeper_groups.copy()
        external_group_copy = external_groups.copy()
        for match_round in range(3):  # 0 - external ID, 1 - name, 2 - reuse groups
            if len(keeper_group_copy) == 0 or len(external_group_copy) == 0:
                break
            if match_round == 0:
                group_lookup = {x.external_id: x for x in keeper_group_copy.values() if x.external_id}
            elif match_round == 1:
                group_lookup = {x.name.casefold(): x for x in keeper_group_copy.values()}
            elif match_round == 2:
                group_lookup = {key: value for key, value in
                                zip(external_group_copy.keys(), keeper_group_copy.values()) if value.external_id}
            else:
                continue

            for group_id in list(external_group_copy.keys()):
                group = external_group_copy[group_id]
                if match_round in (0, 2):
                    key = group.id
                elif match_round == 1:
                    key = group.name.casefold()
                else:
                    continue
                if key in group_lookup:
                    keeper_group = group_lookup[key]
                    op = {
                        'op': 'replace',
                        'value': {}
                    }
                    if keeper_group.external_id != group.id:
                        op['value']['externalId'] = group.id
                    if keeper_group.name != group.name:
                        op['value']['displayName'] = group.name
                    if len(op['value']) > 0:
                        payload = {
                            'schemas': ['urn:ietf:params:scim:api:messages:2.0:PatchOp'],
                            'Operations': [op]
                        }
                        try:
                            ScimPushCommand.patch_scim_resource(
                                f'{scim_url}/Groups', keeper_group.id, token, payload, dry_run)
                            keeper_group.external_id = group.id
                            keeper_group.name = group.name
                            logging.info('SCIM updated group "%s"', group.name)
                        except Exception as e:
                            logging.warning('PATCH group "%s" error: %s', group.name, e)

                    del keeper_group_copy[keeper_group.id]
                    del external_group_copy[group.id]

        if len(external_group_copy) > 0:  # add groups
            for group in external_group_copy.values():
                payload = {
                    'schemas': ['urn:ietf:params:scim:schemas:core:2.0:Group;'],
                    'displayName': group.name,
                    'externalId': group.id
                }
                try:
                    rs = ScimPushCommand.post_scim_resource(f'{scim_url}/Groups', token, payload, dry_run)
                    group_id = rs.get('id')
                    if group_id:
                        keeper_group = ScimGroup()
                        keeper_group.id = group_id
                        keeper_group.external_id = rs.get('externalId')
                        keeper_group.name = rs.get('displayName')
                        keeper_groups[group_id] = keeper_group
                        logging.info('SCIM added group "%s"', group.name)
                except Exception as e:
                    logging.warning('POST group "%s" error: %s', group.name, e)
        external_group_copy.clear()

        if len(keeper_group_copy) > 0:  # delete groups
            destructive = kwargs.get('destructive')
            if not isinstance(destructive, int):
                destructive = 0
            for keeper_group_id in keeper_group_copy:
                keeper_group = keeper_group_copy[keeper_group_id]
                try:
                    if destructive > 0 or keeper_group.external_id:
                        ScimPushCommand.delete_scim_resource(f'{scim_url}/Groups', keeper_group_id, token, dry_run)
                        del keeper_groups[keeper_group_id]
                        logging.info('SCIM deleted group "%s"', keeper_group.name)
                    else:
                        if keeper_group.external_id:
                            logging.info('DELETE group "%s" skipped: "Safe Mode" is enforced', keeper_group.name)
                        else:
                            logging.info('DELETE group "%s" skipped: the group is not controlled by SCIM', keeper_group.name)
                except Exception as e:
                    logging.warning('DELETE group "%s" error: %s', keeper_group.name, e)
        keeper_group_copy.clear()

    @staticmethod
    def sync_users(scim_url, token,
                   keeper_users,
                   external_users,
                   dry_run=False):  # type: (str, str, Dict[str, ScimUser], Dict[str, ScimUser], bool) -> None
        keeper_user_copy = keeper_users.copy()
        external_user_copy = external_users.copy()
        for match_round in range(1):  # 0 - email
            if len(keeper_user_copy) == 0 or len(external_user_copy) == 0:
                break
            if match_round == 0:
                user_lookup = {x.email.casefold(): x for x in keeper_user_copy.values()}  # type: Dict[str, ScimUser]
            else:
                continue

            for user_id in list(external_user_copy.keys()):
                user = external_user_copy[user_id]
                if match_round == 0:
                    key = user.email.casefold()
                else:
                    continue
                if key in user_lookup:
                    keeper_user = user_lookup[key]
                    op = {
                        'op': 'replace',
                        'value': {}
                    }
                    if keeper_user.external_id != user.id:
                        op['value']['externalId'] = user.id
                    if keeper_user.full_name != user.full_name:
                        op['value']['displayName'] = user.full_name
                    if keeper_user.last_name != user.last_name:
                        op['value']['name.familyName'] = user.last_name
                    if keeper_user.first_name != user.first_name:
                        op['value']['name.givenName'] = user.first_name
                    if keeper_user.active != user.active:
                        op['value']['active'] = user.active

                    if len(op['value']) > 0:
                        payload = {
                            'schemas': ['urn:ietf:params:scim:api:messages:2.0:PatchOp'],
                            'Operations': [op]
                        }
                        try:
                            ScimPushCommand.patch_scim_resource(
                                f'{scim_url}/Users', keeper_user.id, token, payload, dry_run)
                            keeper_user.external_id = user.id
                            keeper_user.full_name = user.full_name
                            keeper_user.first_name = user.first_name
                            keeper_user.last_name = user.last_name
                            keeper_user.active = user.active
                            logging.info('SCIM updated user "%s"', user.email)
                        except Exception as e:
                            logging.warning('PATCH user "%s" error: %s', user.email, e)

                    del keeper_user_copy[keeper_user.id]
                    del external_user_copy[user.id]

        if len(external_user_copy) > 0:  # add users
            for user in external_user_copy.values():
                if not user.active:
                    continue

                payload = {
                    'schemas': ["urn:ietf:params:scim:schemas:core:2.0:User",
                                "urn:ietf:params:scim:schemas:extension:enterprise:2.0:User"],
                    'userName': user.email,
                    'externalId': user.id,
                    'displayName': user.full_name or '',
                    'name': {
                        "givenName": user.first_name or '',
                        "familyName": user.last_name or ''
                    },
                    'active': user.active
                }
                try:
                    rs = ScimPushCommand.post_scim_resource(f'{scim_url}/Users', token, payload, dry_run)
                    user_id = rs.get('id')
                    if user_id:
                        keeper_user = ScimUser()
                        keeper_user.id = user_id
                        keeper_user.email = user.email
                        keeper_user.active = user.active
                        keeper_user.external_id = user.id
                        keeper_user.first_name = user.first_name
                        keeper_user.last_name = user.last_name
                        keeper_users[user_id] = keeper_user
                        logging.info('SCIM added user "%s"', user.email)
                except Exception as e:
                    logging.warning('POST email "%s" error: %s', user.email, e)
        external_user_copy.clear()

        if len(keeper_user_copy) > 0:  # delete users
            for keeper_user_id in keeper_user_copy:
                keeper_user = keeper_user_copy[keeper_user_id]
                if not keeper_user.active:
                    continue
                try:
                    ScimPushCommand.delete_scim_resource(f'{scim_url}/Users', keeper_user_id, token, dry_run)
                    del keeper_users[keeper_user_id]
                    logging.info('SCIM deleted user "%s"', keeper_user.email)
                except Exception as e:
                    logging.warning('DELETE user "%s" error: %s', keeper_user.email, e)
        keeper_user_copy.clear()

    @staticmethod
    def sync_membership(scim_url,  # type: str
                        token,  # type: str
                        keeper_groups,   # type: Dict[str, ScimGroup]
                        keeper_users,    # type: Dict[str, ScimUser]
                        external_users,  # type: Dict[str, ScimUser]
                        dry_run=False,   # type: bool
                        **kwargs         # type: Any
                        ):  # type: (...) -> None
        destructive = kwargs.get('destructive')
        if not isinstance(destructive, int):
            destructive = 0

        keeper_user_lookup = {x.email: x for x in keeper_users.values()}   # type: Dict[str, ScimUser]
        keeper_group_map = {x.external_id: x.id for x in keeper_groups.values() if x.external_id}
        for user in external_users.values():
            if user.email not in keeper_user_lookup:
                continue
            keeper_user = keeper_user_lookup[user.email]
            keeper_user_groups = set(keeper_user.groups or [])
            add_groups = []
            remove_groups = []
            for external_group_id in user.groups:
                if external_group_id in keeper_group_map:
                    keeper_group_id = keeper_group_map[external_group_id]
                    if keeper_group_id in keeper_user_groups:
                        keeper_user_groups.remove(keeper_group_id)
                    else:
                        add_groups.append(keeper_group_id)
            if len(keeper_user_groups) > 0:
                if destructive > 0:
                    remove_groups.extend(keeper_user_groups)
                else:
                    for keeper_group_id in keeper_user_groups:
                        if keeper_group_id in keeper_groups:
                            keeper_group = keeper_groups[keeper_group_id]
                            if keeper_group.external_id:
                                remove_groups.append(keeper_group_id)
                            else:
                                logging.info('Remove team \"%s\" user "%s" membership skipped. Team is not controlled by SCIM',
                                             keeper_group.name, user.email)
                        else:
                            logging.info('Remove team Id "%s" from user "%s" skipped. Team is outside of SCIM node', keeper_group_id, user.email)

            if len(add_groups) > 0 or len(remove_groups) > 0:
                payload = {
                    'schemas': ['urn:ietf:params:scim:api:messages:2.0:PatchOp'],
                    'Operations': []
                }
                if len(add_groups) > 0:
                    payload['Operations'].append({
                        'op': 'add',
                        'path': 'groups',
                        'value': [{'value': x} for x in add_groups]
                    })
                if len(remove_groups) > 0:
                    payload['Operations'].append({
                        'op': 'remove',
                        'path': 'groups',
                        'value': [{'value': x} for x in remove_groups]
                    })
                try:
                    ScimPushCommand.patch_scim_resource(
                        f'{scim_url}/Users', keeper_user.id, token, payload, dry_run)
                    logging.info('SCIM changed user "%s" membership: %d added; %d removed',
                                 keeper_user.email, len(add_groups), len(remove_groups))
                except Exception as e:
                    logging.warning('PATCH user "%s" membership error: %s', keeper_user.email, e)

    @staticmethod
    def post_scim_resource(url, token, payload, dry_run=False):
        if dry_run:
            logging.info(f'POST {url}')
            logging.info(json.dumps(payload, indent=2))
            response = payload.copy()
            response['id'] = utils.generate_uid()
            return response
        else:
            headers = {
                'Authorization': f'Bearer {token}',
            }
            rs = requests.post(url, headers=headers, json=payload)
            if rs.status_code >= 300:
                raise CommandError('', f'POST error: {rs.status_code}')
            if rs.status_code in (200, 201):
                return rs.json()

    @staticmethod
    def patch_scim_resource(url, resource_id, token, payload, dry_run=False):
        patch_url = f'{url}/{resource_id}'
        if dry_run:
            logging.info(f'PATCH {patch_url}')
            logging.info(json.dumps(payload, indent=2))
        else:
            headers = {
                'Authorization': f'Bearer {token}',
            }
            rs = requests.patch(patch_url, headers=headers, json=payload)
            if rs.status_code >= 300:
                raise CommandError('', f'PATCH error: {rs.status_code}')
            if rs.status_code == 200:
                return rs.json()

    @staticmethod
    def delete_scim_resource(url, resource_id, token, dry_run=False):
        patch_url = f'{url}/{resource_id}'
        if dry_run:
            logging.info(f'DELETE {patch_url}')
        else:
            headers = {
                'Authorization': f'Bearer {token}',
            }
            rs = requests.delete(patch_url, headers=headers)
            if rs.status_code >= 300:
                raise CommandError('', f'DELETE error: {rs.status_code}')

    @staticmethod
    def get_scim_resource(url, token):
        resources = []
        start_index = 1
        count = 500
        headers = {
            'Authorization': f'Bearer {token}'
        }
        comps = urlparse(url)

        while True:
            q = parse_qsl(comps.query, keep_blank_values=True)
            q.append(('startIndex', str(start_index)))
            q.append(('count', str(count)))
            query = urlencode(q, doseq=True)
            url_comp = (comps.scheme, comps.netloc, comps.path, None, query, None)
            rq_url = urlunparse(url_comp)
            rs = requests.get(rq_url, headers=headers)
            if rs.status_code != 200:
                raise Exception(f'SCIM GET error code "{rs.status_code}"')
            response = rs.json()
            if 'Resources' in response:
                resources.extend(response['Resources'])
            total_results = response['totalResults']
            items_per_page = response['itemsPerPage']
            start_index = response['startIndex'] + items_per_page
            if start_index >= total_results:
                break
        return resources

    @staticmethod
    def scim_keeper(scim_url, token):  # type: (str, str) -> Iterable[Union[ScimUser, ScimGroup]]
        user_resource = ScimPushCommand.get_scim_resource(f'{scim_url}/Users', token)
        group_resource = ScimPushCommand.get_scim_resource(f'{scim_url}/Groups', token)
        for group in group_resource:
            group_id = group.get('id')
            group_name = group.get('displayName')
            if group_id and group_name:
                g = ScimGroup()
                g.id = group_id
                g.name = group_name
                g.external_id = group.get('externalId')
                yield g
        for user in user_resource:
            user_id = user.get('id')
            email = user.get('userName')
            if user_id and email:
                u = ScimUser()
                u.id = user_id
                u.email = email
                u.active = user.get('active') is True
                u.external_id = user.get('externalId')
                u.full_name = user.get('displayName')
                if 'name' in user:
                    name = user['name']
                    u.first_name = name.get('givenName')
                    u.last_name = name.get('familyName')
                if 'groups' in user:
                    for group in user['groups']:
                        if 'value' in group:
                            group_id = group.get('value')
                            if group_id:
                                u.groups.append(group_id)
                yield u


class AdCrmDataSource(ICrmDataSource):
    def __init__(self, ad_url, ad_user, ad_password, scim_groups):  # type: (str, str, str, List[str]) -> None
        super().__init__()
        self.ad_url = ad_url
        self.ad_user = ad_user
        self.ad_password = ad_password
        self.scim_groups = scim_groups

    def populate(self):
        try:
            import ldap3
            from ldap3.utils.conv import escape_filter_chars
        except ModuleNotFoundError:
            raise CommandError('', 'LDAP3 client is not installed.\npip install ldap3')

        server = ldap3.Server(self.ad_url)
        with ldap3.Connection(server, user=self.ad_user, password=self.ad_password,
                              authentication=ldap3.SIMPLE if server.ssl else ldap3.NTLM) as connection:
            connection.bind()
            if not connection.search('', '(class=*)', search_scope=ldap3.BASE, attributes=["*"]):
                raise CommandError('', 'Active Directory: cannot query Root DSE')
            if len(connection.entries) == 0:
                raise CommandError('', 'Active Directory: cannot query Root DSE')
            root_dn = ''
            entry = connection.entries[0]
            entry_attributes = set(entry.entry_attributes)
            if 'rootDomainNamingContext' in entry_attributes:
                root_dn = entry.rootDomainNamingContext.value
            if not root_dn and 'defaultNamingContext' in entry_attributes:
                root_dn = entry.defaultNamingContext.value
            if not root_dn and 'namingContexts' in entry_attributes:
                attrs = entry.namingContexts.values
                if isinstance(attrs, list) and len(attrs) > 0:
                    root_dn = attrs[0]

            scim_groups = {}           # type: Dict[str, ScimGroup]
            for scim_group in self.scim_groups:
                if scim_group.lower().startswith('cn='):
                    rs = connection.extend.standard.paged_search(
                        scim_group, f'(objectClass=group)',
                        search_scope=ldap3.BASE, attributes=['objectGUID', 'name'], generator=False)
                else:
                    rs = connection.extend.standard.paged_search(
                        root_dn, f'(&(objectClass=group)(name={escape_filter_chars(scim_group)}))',
                        search_scope=ldap3.SUBTREE, attributes=['objectGUID', 'name'], generator=False)

                group_entry = next((x for x in rs if x.get('type') == 'searchResEntry'), None)
                if group_entry:
                    t = group_entry.get('type')
                    if t != 'searchResEntry':
                        continue
                    group_dn = group_entry['dn']
                    attrs = group_entry['attributes']
                    scim_group = ScimGroup()
                    scim_group.id = attrs.get('objectGUID')
                    scim_group.name = attrs.get('name')
                    scim_groups[group_dn] = scim_group
                else:
                    self.debug_logger(f'AD Group "{scim_group}" could not be resolved')
                    self._load_errors = True

            if len(scim_groups) == 0:
                raise Exception('No Active Directory groups could be resolved')

            scim_users = {}           # type: Dict[str, ScimUser]

            for group_dn, group in scim_groups.items():
                group_users = connection.extend.standard.paged_search(
                    root_dn, f'(&(objectClass=user)(memberOf={escape_filter_chars(group_dn)}))',
                    search_scope=ldap3.SUBTREE, paged_size=1000, generator=True,
                    attributes=['objectGUID', 'mail', 'userPrincipalName', 'givenName', 'accountExpires',
                                'sn', 'cn', 'memberOf'])
                now = datetime.datetime.now().timestamp()
                for u in group_users:
                    t = u.get('type')
                    if t != 'searchResEntry':
                        continue

                    attrs = u['attributes']
                    if 'objectGUID' in attrs:
                        user_id = attrs['objectGUID']
                    else:
                        continue
                    su = scim_users.get(user_id)
                    if not su:
                        email = ''
                        if 'mail' in attrs:
                            email = attrs['mail']
                        if not email and 'userPrincipalName' in attrs:
                            email = attrs['userPrincipalName']
                        if not email:
                            continue
                        su = ScimUser()
                        su.id = user_id
                        su.email = email
                        if 'cn' in attrs:
                            su.full_name = attrs['cn']
                        if 'givenName' in attrs:
                            su.first_name = attrs['givenName']
                        if 'sn' in attrs:
                            su.last_name = attrs['sn']
                        if 'accountExpires' in attrs:
                            ae = attrs['accountExpires']
                            if isinstance(ae, datetime.datetime):
                                su.active = ae.timestamp() > now
                        scim_users[user_id] = su
                    su.groups.append(group_dn)

            yield from scim_groups.values()
            yield from scim_users.values()


class LocalCrmDataSource(ICrmDataSource):
    USER_FIELDS = ['id', 'external_id', 'email', 'full_name', 'first_name', 'last_name', 'active', 'groups']
    GROUP_FIELDS = ['id', 'external_id', 'name']

    def __init__(self, data):  # type: (Dict[str, Any]) -> None
        super().__init__()
        self.data = data

    def populate(self):      # type: () -> Iterable[Union[ScimGroup, ScimUser]]
        if not isinstance(self.data, dict):
            raise CommandError('', f'SCIM data are missing')

        if 'users' in self.data:
            for user in self.data['users']:
                if not isinstance(user, dict):
                    continue
                su = ScimUser()
                for key, value in user.items():
                    if hasattr(su, key):
                        setattr(su, key, value)
                    else:
                        user_fields = ','.join(self.USER_FIELDS)
                        raise CommandError('', f'SCIM user: unsupported field "{key}": {user_fields}')
                if not su.id or not su.email:
                    raise CommandError('', f'SCIM user: fields "id" and "email" are required')
                yield su

        if 'groups' in self.data:
            for group in self.data['groups']:
                if not isinstance(group, dict):
                    continue
                sg = ScimGroup()
                for key, value in group.items():
                    if hasattr(sg, key):
                        setattr(sg, key, value)
                    else:
                        group_fields = ','.join(self.GROUP_FIELDS)
                        raise CommandError('', f'SCIM group: unsupported field "{key}": {group_fields}')
                if not sg.id or not sg.name:
                    raise CommandError('', f'SCIM group: fields "id" and "name" are required')
                yield sg


class GoogleCrmDataSource(ICrmDataSource):
    def __init__(self, admin_account, credentials, scim_groups):  # type: (str, dict, List[str]) -> None
        super().__init__()
        self.admin_account = admin_account
        self.credentials = credentials
        self.scim_groups = scim_groups

    @staticmethod
    def parse_google_user(scim_user):   # type: (dict) -> Optional[ScimUser]
        if 'id' in scim_user and 'primaryEmail' in scim_user:
            u = ScimUser()
            u.id = scim_user['id']
            u.email = scim_user['primaryEmail']
            u.active = not (scim_user.get('suspended') is True)
            if 'name' in scim_user:
                scim_user_name = scim_user['name']
                if 'givenName' in scim_user_name:
                    u.first_name = scim_user['name']['givenName']
                if 'familyName' in scim_user_name:
                    u.last_name = scim_user['name']['familyName']
            if u.first_name or u.last_name:
                u.full_name = f'{(u.first_name or "")} {(u.last_name or "")}'.strip()
            return u

    @staticmethod
    def parse_google_group(scim_group):  # type: (dict) -> Optional[ScimGroup]
        if 'id' in scim_group and 'name' in scim_group:
            g = ScimGroup()
            g.id = scim_group.get('id')
            g.name = scim_group.get('name')
            return g

    def populate(self) -> Iterable[Union[ScimGroup, ScimUser]]:
        try:
            from google.oauth2 import service_account
            import googleapiclient.discovery
            logging.getLogger('googleapiclient.discovery_cache').setLevel(logging.ERROR)
        except ModuleNotFoundError:
            raise CommandError('', 'Google Cloud client is not installed.\npip install google-api-python-client')

        scopes = ['https://www.googleapis.com/auth/admin.directory.group.readonly',
                  'https://www.googleapis.com/auth/admin.directory.group.member.readonly',
                  'https://www.googleapis.com/auth/admin.directory.user.readonly']
        cred = service_account.Credentials.from_service_account_info(self.credentials, scopes=scopes).with_subject(self.admin_account)
        directory = googleapiclient.discovery.build('admin', 'directory_v1', credentials=cred, static_discovery=False)

        self.debug_logger('Resolving "SCIM Groups" content')
        scim_users = {}    # type: Dict[str, ScimUser]
        scim_groups = {}   # type: Dict[str, ScimGroup]

        if isinstance(self.scim_groups, list):
            for name in self.scim_groups:
                if utils.is_email(name):
                    rs = directory.groups().list(customer='my_customer', query=f'email={name}').execute()
                    groups = rs.get('groups')
                    if isinstance(groups, list) and len(groups) > 0:
                        for group in groups:
                            g = GoogleCrmDataSource.parse_google_group(group)
                            if isinstance(g, ScimGroup):
                                self.debug_logger(f'Found Google group "{g.name}" for email "{name}"')
                                scim_groups[g.id] = g
                    else:
                        rs = directory.users().list(customer='my_customer', query=f'email={name}').execute()
                        users = rs.get('users')
                        if isinstance(users, list) and len(users) > 0:
                            for user in users:
                                u = GoogleCrmDataSource.parse_google_user(user)
                                if isinstance(u, ScimUser):
                                    self.debug_logger(f'Found Google user for email "{name}"')
                                    scim_users[u.id] = u
                        else:
                            self.debug_logger(f'An email "{name}" could not be resolved as either Google User or Group')
                            self._load_errors = True
                else:
                    rs = directory.groups().list(customer='my_customer', query=f'name=\'{name}\'').execute()
                    groups = rs.get('groups')
                    if isinstance(groups, list) and len(groups) > 0:
                        for group in groups:
                            g = GoogleCrmDataSource.parse_google_group(group)
                            if isinstance(g, ScimGroup):
                                self.debug_logger(f'Found Google group "{g.name}" by name')
                                scim_groups[g.id] = g
                    else:
                        self.debug_logger(f'A name "{name}" could not be resolved to Google Group. Names are case sensitive')
                        self._load_errors = True
        if len(scim_groups) == 0 and len(scim_users) == 0:
            raise Exception('no Google Workspace groups could be resolved')

        self.debug_logger('Loading all users')
        user_lookup = {}   # type: Dict[str, ScimUser]
        user_request = directory.users().list(customer='my_customer')
        while user_request:
            user_response = user_request.execute()
            for user in user_response.get('users', []):
                u = GoogleCrmDataSource.parse_google_user(user)
                if isinstance(u, ScimUser):
                    user_lookup[u.id] = u
            user_request = directory.users().list_next(previous_request=user_request, previous_response=user_response)
        self.debug_logger(f'Total {len(user_lookup)} Google user(s) loaded')

        membership_cache = {}   # type: Dict[str, List[str]]
        for g_id, group in scim_groups.items():
            group_ids = [g_id]            # type: List[str]
            queued_ids = set(group_ids)   # type: Set[str]
            pos = 0
            while pos < len(group_ids):
                group_id = group_ids[pos]
                pos += 1
                if group_id not in membership_cache:
                    members_request = directory.members().list(groupKey=group_id)
                    members = []
                    while members_request:
                        try:
                            members_response = members_request.execute()
                            google_members = members_response.get('members')
                            if isinstance(google_members, list):
                                members.extend((x['id'] for x in google_members))
                            else:
                                break
                            members_request = directory.members().list_next(previous_request=members_request, previous_response=members_response)
                        except Exception as e:
                            self.debug_logger(f'Error loading members for group "{group_id}": {e}')
                            break
                    membership_cache[group_id] = members
                for member_id in membership_cache[group_id]:
                    if member_id in user_lookup:
                        u = user_lookup[member_id]
                        u.groups.append(group_id)
                        if u.id not in scim_users:
                            scim_users[u.id] = u
                    else:
                        if member_id not in queued_ids:
                            queued_ids.add(member_id)
                            group_ids.append(member_id)

        yield from scim_groups.values()
        yield from scim_users.values()



import sys
if sys.version_info >= (3, 8):
  import pytest
  from unittest import TestCase
  from keepercommander.service.util.parse_keeper_response import KeeperResponseParser

  class TestKeeperResponseParser(TestCase):
      def test_parse_ls_command(self):
          """Test parsing of 'ls' command output"""
          sample_output = """# Folder UID                           Title                                                   Flags
    1   b4pBzT1WowoUXHk_US0SCg   Root                                                    RS
  # Record UID                           Type           Title                                      Description   
    1   dGJ3xbH8CXhNF00FBX0wMA            login          My Login                                   Important"""
          
          result = KeeperResponseParser._parse_ls_command(sample_output)
          
          self.assertEqual(result['status'], 'success')
          self.assertEqual(result['command'], 'ls')
          self.assertEqual(len(result['data']['folders']), 1)
          self.assertEqual(len(result['data']['records']), 1)
          
          folder = result['data']['folders'][0]
          self.assertEqual(folder['number'], 1)
          self.assertEqual(folder['name'], 'Root')
          
          record = result['data']['records'][0]
          self.assertEqual(record['number'], 1)
          self.assertEqual(record['title'], 'My Login')
          self.assertEqual(record['description'], 'Important')

      def test_parse_tree_command(self):
          """Test parsing of 'tree' command output"""
          sample_output = """Root
  Folder1
    SubFolder1
  Folder2"""
          
          result = KeeperResponseParser._parse_tree_command(sample_output)
          
          self.assertEqual(result['status'], 'success')
          self.assertEqual(result['command'], 'tree')
          self.assertEqual(len(result['data']['tree']), 4)  # Updated: now returns dict with 'tree' key
          
          self.assertEqual(result['data']['tree'][0]['level'], 0)
          self.assertEqual(result['data']['tree'][0]['name'], 'Root')
          self.assertEqual(result['data']['tree'][0]['path'], 'Root')
          
          self.assertEqual(result['data']['tree'][1]['level'], 0)
          self.assertEqual(result['data']['tree'][1]['name'], 'Folder1')
          self.assertEqual(result['data']['tree'][1]['path'], 'Folder1')

      def test_parse_mkdir_command(self):
          """Test parsing of 'mkdir' command output"""

          result = KeeperResponseParser._parse_mkdir_command('b4pBzT1WowoUXHk_US0SCg')
          self.assertEqual(result['data']['folder_uid'], 'b4pBzT1WowoUXHk_US0SCg')
          
          result = KeeperResponseParser._parse_mkdir_command('Created folder with folder_uid=b4pBzT1WowoUXHk_US0SCg')
          self.assertEqual(result['data']['folder_uid'], 'b4pBzT1WowoUXHk_US0SCg')

      def test_parse_get_command(self):
          """Test parsing of 'get' command output"""
          sample_output = """Title: Test Record
  Username: testuser
  Password: testpass
  URL: https://example.com"""
          
          result = KeeperResponseParser._parse_get_command(sample_output)
          
          self.assertEqual(result['status'], 'success')
          self.assertEqual(result['command'], 'get')
          self.assertEqual(result['data']['title'], 'Test Record')
          self.assertEqual(result['data']['username'], 'testuser')
          self.assertEqual(result['data']['password'], 'testpass')
          self.assertEqual(result['data']['url'], 'https://example.com')


from __future__ import annotations
import argparse
from ..discover import PAMGatewayActionDiscoverCommandBase
from ...display import bcolors
from ... import vault
from ...proto import router_pb2
from ...sync_down import sync_down
from keeper_secrets_manager_core.utils import url_safe_str_to_bytes
from ..pam.router_helper import router_set_record_rotation_information
from ...discovery_common.record_link import RecordLink
from ...discovery_common.constants import PAM_USER, PAM_MACHINE, PAM_DATABASE, PAM_DIRECTORY
from ...discovery_common.types import UserAcl, UserAclRotationSettings
import re
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ...vault import TypedRecord
    from ...params import KeeperParams


class PAMDebugRotationSettingsCommand(PAMGatewayActionDiscoverCommandBase):
    parser = argparse.ArgumentParser(prog='pam-action-debug-rotation')

    # The record to base everything on.
    parser.add_argument('--user-record-uid', '-i', required=True, dest='user_record_uid', action='store',
                        help='PAM user record UID.')
    parser.add_argument('--configuration-record-uid', '-c', required=False,
                        dest='configuration_record_uid', action='store', help='PAM configuration record UID.')
    parser.add_argument('--resource-record-uid', '-r', required=False,
                        dest='resource_record_uid',  action='store', help='PAM resource record UID.')
    parser.add_argument('--noop', required=False, dest='noop', action='store_true',
                        help='User is part of a No Operation.')
    parser.add_argument('--force', required=False, dest='force', action='store_true',
                        help='Force reset of the rotation settings.')
    parser.add_argument('--dry-run', required=False, dest='dry_run', action='store_true',
                        help='Do not create or update anything.')

    def get_parser(self):
        return PAMDebugRotationSettingsCommand.parser

    def execute(self, params: KeeperParams, **kwargs):

        user_record_uid = kwargs.get("user_record_uid")
        resource_record_uid = kwargs.get("resource_record_uid")
        configuration_record_uid = kwargs.get("configuration_record_uid")
        noop = kwargs.get("noop", False)
        force = kwargs.get("force", False)
        dry_run = kwargs.get("dry_run", False)

        print("")

        user_record = vault.KeeperRecord.load(params, user_record_uid)  # type: Optional[TypedRecord]
        if user_record is None:
            print(f"{bcolors.FAIL}The PAM user record does not exists.{bcolors.ENDC}")
            return

        if user_record.record_type != PAM_USER:
            print(f"{bcolors.FAIL}The PAM user record is a {PAM_USER}. "
                  f"The record is {user_record.record_type}{bcolors.ENDC}")
            return

        record_rotation = params.record_rotation_cache.get(user_record_uid)
        if record_rotation is None:
            print(f"{bcolors.WARNING}The protobuf rotation settings are missing. Attempting to create.{bcolors.ENDC}")

            if configuration_record_uid is None:
                print(f"{bcolors.FAIL}Cannot determine PAM configuration, please set the "
                      f"-c, --configuration-record-uid parameter for this command.{bcolors.ENDC}")
                return

            configuration_record = vault.KeeperRecord.load(params,
                                                           configuration_record_uid)  # type: Optional[TypedRecord]
            if configuration_record is None:
                print(f"{bcolors.FAIL}Configuration record does not exists.{bcolors.ENDC}")
                return

            if re.search(r'^pam.*Configuration$', configuration_record.record_type) is None:
                print(
                    f"{bcolors.FAIL}The configuration record is not a configuration record. "
                    f"It's {configuration_record.record_type} record.{bcolors.ENDC}")
                return

            if resource_record_uid is None:
                while True:
                    yn = input("The resource record UID was not set. "
                               "This user does not belongs to a machine, database, or directory; "
                               "It's an IAM, Azure, or Domain Controller user? [Y/N]").lower()
                    if yn == "n":
                        print(f"{bcolors.FAIL}Since a resource is needed, please set --resource-record-uid, -r "
                              f"parameter for the this command.{bcolors.ENDC}")
                        return
                    elif yn == "y":
                        break

            if resource_record_uid is not None:

                resource_record = vault.KeeperRecord.load(params,
                                                          resource_record_uid)  # type: Optional[TypedRecord]
                if resource_record is None:
                    print(f"{bcolors.FAIL}The resource record does not exists.{bcolors.ENDC}")
                    return

                if resource_record.record_type not in [PAM_MACHINE, PAM_DATABASE, PAM_DIRECTORY]:
                    print(f"{bcolors.FAIL}The resource is NOT a "
                          f"{PAM_MACHINE}, {PAM_DATABASE}, or {PAM_DIRECTORY} record. "
                          f"It's a {resource_record.record_type}.{bcolors.ENDC}")
                    return

            parent_uid = resource_record_uid or configuration_record_uid

            # Create rotation settings for the pamUser.
            rq = router_pb2.RouterRecordRotationRequest()
            rq.recordUid = url_safe_str_to_bytes(user_record_uid)
            rq.revision = 0
            rq.configurationUid = url_safe_str_to_bytes(configuration_record_uid)
            rq.resourceUid = url_safe_str_to_bytes(parent_uid)
            rq.schedule = ''
            rq.pwdComplexity = b''
            rq.disabled = False

            if dry_run is False:
                router_set_record_rotation_information(params, rq)

                params.sync_data = True
                sync_down(params)

                record_rotation = params.record_rotation_cache.get(user_record_uid)
                if record_rotation is None:
                    print(f"{bcolors.FAIL}Protobuf rotation settings did not create.{bcolors.ENDC}")
                    return
            else:
                print(f"{bcolors.OKBLUE}DRY RUN: Would have created the protobuf rotation settings.{bcolors.ENDC}")
                record_rotation = {
                    "configuration_uid": configuration_record_uid,
                    "resource_uid": resource_record_uid
                }

        configuration_record_uid = record_rotation.get("configuration_uid")
        if configuration_record_uid is None:
            print(f"{bcolors.FAIL}Record does not have the PAM Configuration set.{bcolors.ENDC}")
            return

        print(f"{bcolors.BOLD}Configuration Record UID{bcolors.ENDC}: {configuration_record_uid}")

        configuration_record = vault.KeeperRecord.load(params,
                                                       configuration_record_uid)  # type: Optional[TypedRecord]
        if configuration_record is None:
            print(f"{bcolors.FAIL}Configuration record does not exists.{bcolors.ENDC}")
            return

        resource_record_uid = record_rotation.get("resource_uid")
        if resource_record_uid is not None:

            print(f"{bcolors.BOLD}Resource Record UID{bcolors.ENDC}: {resource_record_uid}")

            resource_record = vault.KeeperRecord.load(params,
                                                      resource_record_uid)  # type: Optional[TypedRecord]
            if resource_record is None:
                print(f"{bcolors.FAIL}The resource record does not exists.{bcolors.ENDC}")
                return

            if resource_record.record_type not in [PAM_MACHINE, PAM_DATABASE, PAM_DIRECTORY]:
                print(f"{bcolors.FAIL}The resource is a {PAM_MACHINE}, {PAM_DATABASE}, or {PAM_DIRECTORY} record. "
                      f"It's a {resource_record.record_type}.{bcolors.ENDC}")
                return

        record_link = RecordLink(record=configuration_record, params=params)

        parent_uid = resource_record_uid or configuration_record_uid
        parent_vertex = record_link.get_record_link(parent_uid)
        if parent_vertex is None:
            parent_type = "configuration"
            if resource_record_uid is not None:
                parent_type = "resource"
            print(f"{bcolors.FAIL}Could not find the parent linking vertex for the {parent_type}.{bcolors.ENDC}")
            return

        print(f"{bcolors.BOLD}User Record UID{bcolors.ENDC}: {user_record_uid}")

        user_vertex = record_link.get_record_link(user_record_uid)
        if user_vertex is None:
            print(f"{bcolors.WARNING}The user vertex is missing; creating.{bcolors.ENDC}")
            record_link.dag.add_vertex(uid=user_record_uid)

        user_acl = record_link.get_acl(user_record_uid, parent_uid)
        if user_acl is None:
            print(f"{bcolors.WARNING}No ACL exists between the user and the parent; creating.{bcolors.ENDC}")
            user_acl = UserAcl.default()
            user_acl.belongs_to = True

        print("")
        if user_acl.rotation_settings is not None:
            if (force is False and (
                    user_acl.rotation_settings.schedule != ""
                    or user_acl.rotation_settings.pwd_complexity != ""
                    or (user_acl.rotation_settings.saas_record_uid_list is not None
                        and len(user_acl.rotation_settings.saas_record_uid_list) != 0))):
                print(f"{bcolors.FAIL}{user_acl.model_dump_json(indent=4)}{bcolors.ENDC}")
                print(f"{bcolors.FAIL}Rotation settings exist in graph, use --force to reset.{bcolors.ENDC}")
                return

        # Reset the rotation settings.
        user_acl.rotation_settings = UserAclRotationSettings()
        user_acl.rotation_settings.noop = noop
        if resource_record_uid is None:
            user_acl.is_iam_user = True

        # Connect the user to the parent (configuration or resource)
        record_link.belongs_to(user_record_uid, parent_uid, acl=user_acl)

        # If parent is not a configuration, make sure there is a LINK from the resource to the configuration.
        if parent_uid != configuration_record_uid:
            if record_link.get_parent_record_uid(parent_uid) is None:
                print(f"{bcolors.WARNING}Resource record has no LINK to configuration record; "
                      f"creating.{bcolors.ENDC}")
                record_link.belongs_to(configuration_record_uid, parent_uid)

        if dry_run is False:
            record_link.save()

            print(f"{bcolors.OKGREEN}{user_acl.model_dump_json(indent=4)}{bcolors.ENDC}")
            print(f"{bcolors.OKGREEN}Updated the ACL for the user.{bcolors.ENDC}")
        else:
            print(f"{bcolors.OKBLUE}DRY RUN: Would have created this ACL.")
            print(f"{bcolors.OKBLUE}{user_acl.model_dump_json(indent=4)}{bcolors.ENDC}")



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2023 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import json
from typing import Iterable, Dict, List

from . import importer
from .. import vault, record_types
from ..commands.record_edit import RecordEditMixin
from ..params import KeeperParams


def import_to_typed_records(params, records):   # type: (KeeperParams, Iterable[importer.Record]) -> Iterable[vault.TypedRecord]
    record_type_fields = {}    # type: Dict[str, List]
    if params.record_type_cache:
        for rts in params.record_type_cache.values():
            try:
                rto = json.loads(rts)
                if '$id' in rto and 'fields' in rto:
                    record_type_fields[rto['$id'].lower()] = rto['fields']
            except:
                pass

    for record in records:
        if not isinstance(record, importer.Record):
            continue
        record_type = record.type or 'login'
        record_type_lower = record_type.lower()
        if record_type_lower not in record_type_fields:
            record_type_lower = 'login'

        typed_record = vault.TypedRecord()
        typed_record.title = record.title
        typed_record.notes = record.notes
        typed_record.type_name = record_type
        RecordEditMixin.adjust_typed_record_fields(typed_record, record_type_fields[record_type_lower])
        if record.login:
            field = typed_record.get_typed_field('login')
            if not field:
                field = vault.TypedField.new_field('login', '')
                typed_record.custom.append(field)
            field.value = [record.login]
        if record.password:
            field = typed_record.get_typed_field('password')
            if not field:
                field = vault.TypedField.new_field('password', '')
                typed_record.custom.append(field)
            field.value = [record.password]
        if record.login_url:
            field = typed_record.get_typed_field('url')
            if not field:
                field = vault.TypedField.new_field('url', '')
                typed_record.custom.append(field)
            field.value = [record.login_url]

        for f in record.fields:
            if f.type not in record_types.RecordFields:
                f.type = 'text'

            field = typed_record.get_typed_field(f.type, f.label)
            if not field:
                field = vault.TypedField.new_field(f.type, [], f.label)
                typed_record.custom.append(field)
            if isinstance(f.value, list):
                field.value.extend(f.value)
            elif f.value:
                field.value.append(f.value)

        yield typed_record



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2025 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

"""
AAGUID (Authenticator Attestation GUID) related constants and utilities.
These constants help identify the specific authenticator/provider used for WebAuthn credentials.
"""

# AAGUID to provider name mapping
# Based on community-sourced data from https://github.com/passkeydeveloper/passkey-authenticator-aaguids
AAGUID_PROVIDER_MAPPING = {
    'ea9b8d66-4d01-1d21-3ce4-b6b48cb575d4': 'Google Password Manager',
    'adce0002-35bc-c60a-648b-0b25f1f05503': 'Chrome on Mac',
    'fbfc3007-154e-4ecc-8c0b-6e020557d7bd': 'iCloud Keychain',
    'dd4ec289-e01d-41c9-bb89-70fa845d4bf2': 'iCloud Keychain (Managed)',
    '08987058-cadc-4b81-b6e1-30de50dcbe96': 'Windows Hello',
    '9ddd1817-af5a-4672-a2b9-3e3dd95000a9': 'Windows Hello',
    '6028b017-b1d4-4c02-b4b3-afcdafc96bb2': 'Windows Hello',
    '00000000-0000-0000-0000-000000000000': 'Platform Authenticator'
}


def get_provider_name_from_aaguid(aaguid: str) -> str:
    """Get friendly provider name from AAGUID"""
    if not aaguid:
        return None
    
    # Normalize AAGUID format (ensure lowercase, with dashes)
    normalized_aaguid = aaguid.lower()
    if len(normalized_aaguid) == 32:  # No dashes
        normalized_aaguid = f"{normalized_aaguid[:8]}-{normalized_aaguid[8:12]}-{normalized_aaguid[12:16]}-{normalized_aaguid[16:20]}-{normalized_aaguid[20:]}"
    
    return AAGUID_PROVIDER_MAPPING.get(normalized_aaguid)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import os
from dataclasses import dataclass
from typing import Optional
from dotenv import load_dotenv, set_key, dotenv_values
from ..decorators.logging import logger
from .terminal_handler import TerminalHandler
from ... import utils

@dataclass
class ProcessInfo:
    pid: Optional[int]
    terminal: Optional[str]
    is_running: bool
    ngrok_pid: Optional[int] = None
    cloudflare_pid: Optional[int] = None
    
    _env_file = utils.get_default_path() / ".service.env"
    
    @classmethod
    def _str_to_bool(cls, value: str) -> bool:
        return value.lower() in ('true', '1', 'yes', 'on')
    
    @classmethod
    def save(cls, pid, is_running: bool, ngrok_pid: Optional[int] = None, cloudflare_pid: Optional[int] = None) -> None:
        """Save current process information to .env file."""
        
        env_path = str(cls._env_file)
        
        # Create the file if it doesn't exist
        if not cls._env_file.exists():
            cls._env_file.touch()
        
        process_info = {
            'KEEPER_SERVICE_PID': str(pid),
            'KEEPER_SERVICE_TERMINAL': TerminalHandler.get_terminal_info() or '',
            'KEEPER_SERVICE_IS_RUNNING': str(is_running).lower()
        }
        
        if ngrok_pid is not None:
            process_info['KEEPER_SERVICE_NGROK_PID'] = str(ngrok_pid)
        
        if cloudflare_pid is not None:
            process_info['KEEPER_SERVICE_CLOUDFLARE_PID'] = str(cloudflare_pid)
        
        try:
            for key, value in process_info.items():
                set_key(env_path, key, value, quote_mode='never')
            logger.debug("Process information saved successfully to .env")
        except Exception as e:
            logger.error(f"Failed to save process information: {e}")
            pass

    @classmethod
    def load(cls) -> 'ProcessInfo':
        """Load process information from .env file."""
        try:
            for key in dotenv_values(cls._env_file).keys():
                os.environ.pop(key, None)  # Remove old values
            
            if ProcessInfo._env_file.exists():
                load_dotenv(ProcessInfo._env_file, override=True)
                
                pid_str = os.getenv('KEEPER_SERVICE_PID')
                pid = int(pid_str) if pid_str else None
                
                terminal = os.getenv('KEEPER_SERVICE_TERMINAL') or None
                
                is_running_str = os.getenv('KEEPER_SERVICE_IS_RUNNING', 'false')
                is_running = ProcessInfo._str_to_bool(is_running_str)
                
                ngrok_pid_str = os.getenv('KEEPER_SERVICE_NGROK_PID')
                ngrok_pid = int(ngrok_pid_str) if ngrok_pid_str else None
                
                cloudflare_pid_str = os.getenv('KEEPER_SERVICE_CLOUDFLARE_PID')
                cloudflare_pid = int(cloudflare_pid_str) if cloudflare_pid_str else None
                
                logger.debug("Process information loaded successfully from .env")
                return ProcessInfo(
                    pid=pid,
                    terminal=terminal,
                    is_running=is_running,
                    ngrok_pid=ngrok_pid,
                    cloudflare_pid=cloudflare_pid
                )
        except Exception as e:
            logger.error(f"Failed to load process information: {e}")
            pass
        
        return ProcessInfo(pid=None, terminal=None, is_running=False, ngrok_pid=None, cloudflare_pid=None)
    
    @classmethod
    def clear(cls) -> None:
        """Remove the process information file."""
        try:
            cls._env_file.unlink(missing_ok=True)
            logger.debug("Process information file removed")
        except Exception as e:
            logger.error(f"Failed to remove process information file: {e}")
            pass


import json
import os
import io
from typing import Union

from unittest import TestCase, mock

from data_vault import get_synced_params, VaultEnvironment
from helper import KeeperApiHelper

from keepercommander import api, utils, crypto, attachment, vault
from keepercommander.commands import record, record_edit
from keepercommander.error import CommandError


class TestRecord(TestCase):
    vault_env = VaultEnvironment()
    expected_commands = []

    def setUp(self):
        self.communicate_mock = mock.patch('keepercommander.api.communicate').start()
        self.communicate_mock.side_effect = KeeperApiHelper.communicate_command
        self.communicate_mock = mock.patch('keepercommander.api.communicate_rest').start()
        self.communicate_mock.side_effect = TestRecord.communicate_rest_success
        TestRecord.expected_commands.clear()

    def tearDown(self):
        mock.patch.stopall()

    def test_parse_field(self):
        prf = record_edit.RecordEditMixin.parse_field('text.aaa==bbb=ccc')
        self.assertEqual(prf.type, 'text')
        self.assertEqual(prf.label, 'aaa=bbb')
        self.assertEqual(prf.value, 'ccc')

        prf = record_edit.RecordEditMixin.parse_field('aaa==bbb==ccc=')
        self.assertEqual(prf.type, '')
        self.assertEqual(prf.label, 'aaa=bbb=ccc')
        self.assertEqual(prf.value, '')

        prf = record_edit.RecordEditMixin.parse_field('aaa====bbb= =ccc=')
        self.assertEqual(prf.type, '')
        self.assertEqual(prf.label, 'aaa==bbb')
        self.assertEqual(prf.value, '=ccc=')

    def test_create_record(self):
        params = get_synced_params()
        r = vault.KeeperRecord.create(params, 'login')

    def test_add_command(self):
        params = get_synced_params()
        cmd = record_edit.RecordAddCommand()

        with mock.patch('keepercommander.api.sync_down'), \
                mock.patch('keepercommander.record_management.add_record_to_folder') as ar:

            added_record = None     # type: Union[vault.PasswordRecord, vault.TypedRecord, None]
            def artf(p, r, f):
                nonlocal added_record
                added_record = r
                added_record.record_uid = utils.generate_uid()
            ar.side_effect = artf

            with self.assertRaises(CommandError):
                cmd.execute(params, force=True, title='New Record')

            added_record = None
            self.assertRaises(CommandError, cmd.execute, params, force=True, title='New Record')
            # cmd.execute(params, force=True, title='New Record', record_type='legacy',
            #             fields=['login=user@company.com', 'password=password', 'url=https://google.com/', 'AAA=BBB'])
            # self.assertIsNotNone(added_record)
            # self.assertEqual('New Record', added_record.title)
            # self.assertIsInstance(added_record, vault.PasswordRecord)
            # self.assertEqual(added_record.login, 'user@company.com')
            # self.assertEqual(added_record.password, 'password')
            # self.assertEqual(added_record.link, 'https://google.com/')
            # self.assertEqual(len(added_record.custom), 1)
            # value = added_record.get_custom_value('AAA')
            # self.assertEqual(value, 'BBB')

            added_record = None
            cmd.execute(params, force=True, title='New Record', record_type='login',
                        fields=['login=user@company.com', 'password=password', 'url=https://google.com/', 'AAA=BBB'])
            self.assertIsNotNone(added_record)
            self.assertEqual('New Record', added_record.title)
            self.assertIsInstance(added_record, vault.TypedRecord)
            field = added_record.get_typed_field('login')
            self.assertIsNotNone(field)
            self.assertEqual(field.get_default_value(str), 'user@company.com')
            field = added_record.get_typed_field('password')
            self.assertIsNotNone(field)
            self.assertEqual(field.get_default_value(str), 'password')
            field = added_record.get_typed_field('url')
            self.assertIsNotNone(field)
            self.assertEqual(field.get_default_value(str), 'https://google.com/')
            field = added_record.get_typed_field('text', 'AAA')
            self.assertIsNotNone(field)
            self.assertEqual(field.get_default_value(str), 'BBB')

    def test_remove_command_from_root(self):
        params = get_synced_params()
        cmd = record.RecordRemoveCommand()

        record_uid = next(iter(params.subfolder_record_cache['']))
        rec = api.get_record(params, record_uid)

        def pre_delete_command(rq):
            self.assertEqual(rq['command'], 'pre_delete')
            return {
                'pre_delete_response': {
                    'would_delete': {
                        'deletion_summary': ['delete all']
                    },

                    'pre_delete_token': 'token'
                }
            }

        with mock.patch('keepercommander.commands.base.user_choice') as choice_mock:
            choice_mock.return_value = KeyboardInterrupt()
            KeeperApiHelper.communicate_expect([pre_delete_command, 'delete'])
            cmd.execute(params, force=True, records=[rec.record_uid])
            self.assertTrue(KeeperApiHelper.is_expect_empty())

            KeeperApiHelper.communicate_expect([pre_delete_command, 'delete'])
            cmd.execute(params, force=True, record=rec.title)
            self.assertTrue(KeeperApiHelper.is_expect_empty())

            with mock.patch('builtins.print'):
                choice_mock.return_value = 'y'
                KeeperApiHelper.communicate_expect([pre_delete_command, 'delete'])
                cmd.execute(params, records=[rec.record_uid])
                self.assertTrue(KeeperApiHelper.is_expect_empty())

                KeeperApiHelper.communicate_expect([pre_delete_command, 'delete'])
                cmd.execute(params, records=[rec.title])
                self.assertTrue(KeeperApiHelper.is_expect_empty())

                choice_mock.return_value = 'n'
                KeeperApiHelper.communicate_expect([pre_delete_command])
                cmd.execute(params, records=[rec.record_uid])
                self.assertTrue(KeeperApiHelper.is_expect_empty())

                KeeperApiHelper.communicate_expect([pre_delete_command])
                cmd.execute(params, records=[rec.title])
                self.assertTrue(KeeperApiHelper.is_expect_empty())

    def test_search_command(self):
        params = get_synced_params()
        cmd = record.SearchCommand()

        with mock.patch('builtins.print'):
            cmd.execute(params, pattern='.*')
            cmd.execute(params, pattern='Non-existing-name')

    def test_record_list_command(self):
        params = get_synced_params()
        cmd = record.RecordListCommand()

        with mock.patch('builtins.print') as mock_print:
            cmd.execute(params, pattern='record')
            mock_print.assert_called()

            mock_print.reset_mock()
            cmd.execute(params, pattern='INVALID')
            mock_print.assert_not_called()

    def test_shared_list_command(self):
        params = get_synced_params()
        cmd = record.RecordListSfCommand()

        with mock.patch('builtins.print') as mock_print:
            cmd.execute(params, pattern='folder')
            mock_print.assert_called()

            mock_print.reset_mock()
            cmd.execute(params, pattern='INVALID')
            mock_print.assert_not_called()

    def test_team_list_command(self):
        params = get_synced_params()
        cmd = record.RecordListTeamCommand()

        TestRecord.expected_commands.extend(['get_share_objects'])
        with mock.patch('builtins.print'):
            cmd.execute(params)
        self.assertTrue(len(TestRecord.expected_commands) == 0)

    def test_get_record_uid(self):
        params = get_synced_params()
        cmd = record.RecordGetUidCommand()

        record_uid = next(iter(params.subfolder_record_cache['']))
        with mock.patch('builtins.print'), mock.patch('keepercommander.api.get_record_shares'):
            cmd.execute(params, uid=record_uid)
            cmd.execute(params, format='json', uid=record_uid)

    def test_record_list_command_with_fields(self):
        params = get_synced_params()
        cmd = record.RecordListCommand()

        with mock.patch('builtins.print') as mock_print:
            cmd.execute(params, field=['title'], pattern='Record 3')
            printed_args = mock_print.call_args[0][0] if mock_print.call_args else ''
            self.assertIn('Record 3', printed_args)

        with mock.patch('builtins.print') as mock_print:
            cmd.execute(params, field=['title'], pattern='NonExistentRecordName')
            mock_print.assert_not_called()

    def test_get_shared_folder_uid(self):
        params = get_synced_params()
        cmd = record.RecordGetUidCommand()

        shared_folder_uid = next(iter(params.shared_folder_cache))
        with mock.patch('builtins.print'):
            cmd.execute(params, uid=shared_folder_uid)
            cmd.execute(params, format='json', uid=shared_folder_uid)

    def test_get_team_uid(self):
        params = get_synced_params()
        cmd = record.RecordGetUidCommand()

        team_uid = next(iter(params.team_cache))
        with mock.patch('builtins.print'):
            cmd.execute(params, uid=team_uid)
            cmd.execute(params, format='json', uid=team_uid)

    def test_get_invalid_uid(self):
        params = get_synced_params()
        cmd = record.RecordGetUidCommand()

        with self.assertRaises(CommandError):
            cmd.execute(params, uid='invalid')

    def test_append_notes_command(self):
        params = get_synced_params()
        cmd = record_edit.RecordAppendNotesCommand()

        record_uid = next(iter(params.subfolder_record_cache['']))
        with mock.patch('keepercommander.record_management.update_record'):
            cmd.execute(params, notes='notes', record=record_uid)

            with mock.patch('builtins.input', return_value='data'):
                cmd.execute(params, record=record_uid)

        with self.assertRaises(CommandError):
            cmd.execute(params, notes='notes', record='invalid')

    def test_download_attachment_command(self):
        params = get_synced_params()
        cmd = record_edit.RecordDownloadAttachmentCommand()

        records = [x for x in params.record_cache.values() if 'extra_unencrypted' in x and len(x['extra_unencrypted']) > 10]
        rec = records[0]
        record_uid = rec['record_uid']
        extra = json.loads(rec['extra_unencrypted'].decode('utf-8'))
        attachments = {}  # type: dict[any, tuple[attachment.AttachmentDownloadRequest, bytes]]
        for file in extra['files']:
            atta_id = file['id']   # type: str
            key = utils.base64_url_decode(file['key'])
            body_encoded = crypto.encrypt_aes_v1(os.urandom(file['size']), key)
            rq = attachment.AttachmentDownloadRequest()
            rq.title = 'Title'
            rq.url = f'https://keepersecurity.com/files/{atta_id}'
            rq.is_gcm_encrypted = False
            rq.encryption_key = key

            attachments[atta_id] = (rq, body_encoded)

        def prepare_download(params, record_uid):
            return (x[0] for x in attachments.values())

        def requests_get(url, **kwargs):
            body = next((x[1] for x in attachments.values() if x[0].url == url), None)
            if not body:
                raise Exception(f'URL \"{url}\" not found.')
            rs = mock.Mock()
            rs.status_code = 200
            stream = io.BytesIO(body)
            rs.raw = stream
            rs.__enter__ = mock.Mock(return_value=rs)
            rs.__exit__ = mock.Mock(return_value=None)
            return rs

        with mock.patch('keepercommander.attachment.prepare_attachment_download', side_effect=prepare_download), \
                mock.patch('requests.get', side_effect=requests_get), \
                mock.patch('builtins.open', mock.mock_open()), \
                mock.patch('os.path.abspath', return_value='/file_name'):
            cmd.execute(params, record=record_uid)

    def test_delete_attachment_command(self):
        params = get_synced_params()
        record_uid = next((x['record_uid'] for x in params.record_cache.values()
                           if 'extra_unencrypted' in x and len(x['extra_unencrypted']) > 10), None)

        rec = vault.KeeperRecord.load(params, record_uid)
        self.assertIsNotNone(rec)
        self.assertIsInstance(rec, vault.PasswordRecord)
        self.assertGreater(len(rec.attachments), 0)

        KeeperApiHelper.communicate_expect(['record_update'])
        cmd = record_edit.RecordDeleteAttachmentCommand()
        cmd.execute(params, name=[rec.attachments[0].id], record=rec.title)
        self.assertTrue(KeeperApiHelper.is_expect_empty())


    @staticmethod
    def communicate_rest_success(params, request, endpoint, **kwargs):
        if 'rs_type' in kwargs:
            rs = kwargs['rs_type']()
        else:
            rs = None

        _, _, command = endpoint.rpartition('/')

        cmd = TestRecord.expected_commands.pop(0)
        if cmd == command:
            return rs

        raise Exception()

